from enum import Enum
from numbers import Number
from typing import Any, Hashable, Iterable, Mapping, Optional, Tuple, Union

import dask.array as da
import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset
import xarray as xr

from ..utils.decorators import geokube_logging
from ..utils.hcube_logger import HCubeLogger
from ..utils.attrs_encoding import CFAttributes
from ..utils.serialization import maybe_convert_to_json_serializable
from .bounds import Bounds, Bounds1D, BoundsND
from .axis import Axis, AxisType
from .enums import LatitudeConvention, LongitudeConvention
from .unit import Unit
from .variable import Variable


class CoordinateType(Enum):
    SCALAR = "scalar"
    DEPENDENT = "dependent"  # equivalent to CF AUXILIARY Coordinate
    INDEPENDENT = "independent"  # equivalent to CF DIMENSION Coordinate


# NOTE: coordinate is a dimension or axis with data and units
# NOTE: coordinate name is dimension/axis name
# NOTE: coordinate axis type is dimension/axis type

FREQ_CODES = {
    "T": "minute",
    "H": "hour",
    "D": "day",
    "M": "month",
    "Y": "year",
    "S": "second",
    "L": "millisecond",
    "U": "microsecond",
    "N": "nanosecond",
    "NS": "nanosecond",
}


class Coordinate(Variable, Axis):
    __slots__ = ("_bounds",)

    _LOG = HCubeLogger(name="Coordinate")

    def __init__(
        self,
        data: Union[np.ndarray, da.Array, xr.Variable],
        axis: Union[str, Axis],
        dims: Optional[Tuple[Axis]] = None,
        units: Optional[Union[Unit, str]] = None,
        bounds: Optional[
            Union[Bounds, np.ndarray, da.Array, xr.Variable]
        ] = None,
        properties: Optional[Mapping[Hashable, str]] = None,
        encoding: Optional[Mapping[Hashable, str]] = None,
    ):
        if data is None:
            raise ValueError("`data` cannot be `None`")
        if not isinstance(axis, (Axis, str)):
            raise TypeError(
                "Expected argument is one of the following types"
                f" `geokube.Axis` or `str`, but provided {type(data)}"
            )
        Axis.__init__(self, name=axis, is_dim=dims is None)
        # We need to update as when calling constructor of Variable, encoding will be overwritten
        if encoding is not None:
            # encoding stored in axis
            self.encoding.update(encoding)
        if (
            not self.is_dim
            and dims is None
            and not isinstance(data, Number)
            and not hasattr(data, "dims")
        ):
            raise ValueError(
                "If coordinate is not a dimension, you need to supply `dims`"
                " argument!",
            )
        if self.is_dim:
            if isinstance(dims, (list, tuple)):
                dims_names = [Axis.get_name_for_object(o) for o in dims]
                dims_tuple = tuple(dims_names)
            elif isinstance(dims, str):
                dims_tuple = (Axis.get_name_for_object(dims),)
            else:
                dims_tuple = ()
            if dims is None or len(dims_tuple) == 0:
                dims = (self.name,)
            else:
                if dims is not None and len(dims_tuple) > 1:
                    raise ValueError(
                        "If the Coordinate is a dimension, it has to depend"
                        f" only on itself, but provided `dims` are: {dims}",
                    )
                if len(dims_tuple) == 1 and dims_tuple[0] != self.name:
                    raise ValueError(
                        "`dims` parameter for dimension coordinate should"
                        " have the same name as axis name!"
                    )
        Variable.__init__(
            self,
            data=data,
            dims=dims,
            units=units if units is not None else self.default_unit,
            properties=properties,
            encoding=self.encoding,
        )
        # Coordinates are always stored as NumPy data
        # import pdb;pdb.set_trace()
        self._data = np.array(self._data)
        self.bounds = bounds
        self._update_properties_and_encoding()

    def __hash__(self):
        # NOTE: maybe hash for Cooridnate should be more complex.
        return Axis.__hash__(self)

    def __eq__(self, other):
        # NOTE: it doesn't take into account real values at all
        return Axis.__eq__(self, other)

    def __ne__(self, other):
        return not self == other

    def _update_properties_and_encoding(self):
        if CFAttributes.STANDARD_NAME.value not in self.properties:
            self.properties[
                CFAttributes.STANDARD_NAME.value
            ] = self.axis_type.axis_type_name
        if CFAttributes.NETCDF_NAME.value not in self.encoding:
            self.encoding[CFAttributes.NETCDF_NAME.value] = self.ncvar
        Coordinate._handle_fill_value_encoding(self)

    @classmethod
    def _handle_fill_value_encoding(cls, obj):
        # NOTE: _FillValue is not applicable for Coordinate
        # as it shouldn't contain missing data
        # To avoid implicit addition of _FillValue encoding
        # it needs to be set to False for netcdf4 engine
        # see https://github.com/pydata/xarray/issues/1598
        if hasattr(obj, "encoding"):
            obj.encoding[CFAttributes.FILL_VALUE.value] = None

    @classmethod
    @geokube_logging
    def _process_bounds(cls, bounds, name, variable_shape, units, axis):
        if bounds is None:
            return None
        if isinstance(bounds, dict):
            if len(bounds) > 0:
                _bounds = {}
            for k, v in bounds.items():
                if isinstance(v, pd.core.indexes.datetimes.DatetimeIndex):
                    v = np.array(v)
                if isinstance(v, Bounds):
                    bound_class = Coordinate._get_bounds_cls(
                        v.shape, variable_shape
                    )
                    _bounds[k] = v
                if isinstance(v, Variable):
                    bound_class = Coordinate._get_bounds_cls(
                        v.shape, variable_shape
                    )
                    _bounds[k] = bound_class(data=v)
                elif isinstance(v, (np.ndarray, da.Array)):
                    # in this case when only a numpy array is passed
                    # we assume 2-D numpy array with shape(coord.dim, 2)
                    #
                    bound_class = Coordinate._get_bounds_cls(
                        v.shape, variable_shape
                    )
                    _bounds[k] = bound_class(
                        data=v,
                        units=units,
                        dims=(axis, Axis("bounds", AxisType.GENERIC)),
                    )
                else:
                    raise TypeError(
                        "Each defined bound is expected to be one of the"
                        " following types `geokube.Variable`, `numpy.array`,"
                        f" or `dask.Array`, but provided {type(bounds)}"
                    )
        elif isinstance(bounds, Bounds):
            bound_class = Coordinate._get_bounds_cls(
                bounds.shape, variable_shape
            )
            _bounds = {f"{name}_bounds": bounds}
        elif isinstance(bounds, Variable):
            bound_class = Coordinate._get_bounds_cls(
                bounds.shape, variable_shape
            )
            _bounds = {f"{name}_bounds": bound_class(bounds)}
        elif isinstance(bounds, (np.ndarray, da.Array)):
            bound_class = Coordinate._get_bounds_cls(
                bounds.shape, variable_shape
            )
            _bounds = {
                f"{name}_bounds": bound_class(
                    data=bounds,
                    units=units,
                    dims=(axis, Axis("bounds", AxisType.GENERIC)),
                )
            }
        else:
            raise TypeError(
                "Expected argument is one of the following types `dict`,"
                " `numpy.ndarray`, or `geokube.Variable`, but provided"
                f" {type(bounds)}"
            )
        return _bounds

    @classmethod
    def _is_valid_1d_bounds(cls, provided_bnds_shape, provided_data_shape):
        ndim = len(provided_bnds_shape) - 1
        if (
            2 * ndim == 2
            and provided_bnds_shape[-1] == 2
            and provided_bnds_shape[0] == provided_data_shape[0]
        ):
            return True
        if (
            provided_data_shape == ()
            and ndim == 0
            and provided_bnds_shape[0] == 2
        ):
            # The case where there is a scalar coordinate with bounds, e.g.
            # after single value selection
            return True
        return False

    @classmethod
    def _is_valid_nd_bounds(cls, provided_bnds_shape, provided_data_shape):
        ndim = len(provided_bnds_shape) - 1
        if (
            provided_bnds_shape[-1] == 2 * ndim
            and tuple(provided_bnds_shape[:-1]) == provided_data_shape
        ):
            return True
        if (
            len(provided_bnds_shape) == 2
            and len(provided_data_shape) == 1
            and provided_bnds_shape[0] == provided_data_shape[0]
            and (provided_bnds_shape[1] == 2 or provided_bnds_shape[1] == 4)
        ):
            # The case of points domain
            return True
        return False

    @classmethod
    @geokube_logging
    def _get_bounds_cls(cls, provided_bnds_shape, provided_data_shape):
        if cls._is_valid_1d_bounds(provided_bnds_shape, provided_data_shape):
            return Bounds1D
        if cls._is_valid_nd_bounds(provided_bnds_shape, provided_data_shape):
            return BoundsND
        raise ValueError(
            "Bounds should have dimensions: (2,), (N,2), (N,M,4), (N,M,L,6),"
            f" ... Provided shape is `{provided_bnds_shape}`"
        )

    @property
    def is_dimension(self) -> bool:
        return super().is_dim

    @property
    def is_independent(self) -> bool:
        return self.is_dimension or self.type is CoordinateType.SCALAR

    @property
    def is_dependent(self) -> bool:
        return not self.is_independent

    @property
    def type(self):
        # Cooridnate is scalar if data shows so. Dim(s) --  always defined
        if self.is_dimension:
            if self.shape == () or self.shape == (1,):
                return CoordinateType.SCALAR
            else:
                return CoordinateType.INDEPENDENT
        elif self.shape == ():
            return CoordinateType.SCALAR
        else:
            return CoordinateType.DEPENDENT

    @property
    def axis_type(self):
        return self._type

    @property
    def bounds(self):
        return self._bounds

    @bounds.setter
    def bounds(self, value):
        self._bounds = Coordinate._process_bounds(
            value,
            name=self.name,
            variable_shape=self.shape,
            units=self.units,
            axis=(Axis)(self),
        )
        if self._bounds is not None:
            self.encoding["bounds"] = next(iter(self.bounds))

    @property
    def has_bounds(self) -> bool:
        return self._bounds is not None

    @property
    # TODO:  check! I think this works only if lat/lon are independent!
    def convention(
        self,
    ) -> Optional[Union[LatitudeConvention, LongitudeConvention]]:
        if self.axis_type is AxisType.LATITUDE:
            return (
                LatitudeConvention.POSITIVE_TOP
                if self.first() > self.last()
                else LatitudeConvention.NEGATIVE_TOP
            )
        if self.axis_type is AxisType.LONGITUDE:
            return (
                LongitudeConvention.POSITIVE_WEST
                if self.min() >= 0
                else LongitudeConvention.NEGATIVE_WEST
            )

    @geokube_logging
    def to_xarray(
        self, encoding=True
    ) -> xr.core.coordinates.DatasetCoordinates:
        var = Variable.to_xarray(self, encoding=encoding)
        Coordinate._handle_fill_value_encoding(var)
        res_name = self.ncvar if encoding else self.name
        dim_names = self.dim_ncvars if encoding else self.dim_names
        da = xr.DataArray(
            var, name=res_name, coords={res_name: var}, dims=dim_names
        )[res_name]
        if self.has_bounds:
            bounds = {
                k: xr.DataArray(
                    Variable.to_xarray(b, encoding=encoding), name=k
                )
                for k, b in self.bounds.items()
            }
            da.encoding["bounds"] = " ".join(bounds.keys())
        else:
            bounds = {}
        return xr.Dataset(coords={da.name: da, **bounds})

    def to_dict(self, unique_values=False):
        axis_specific_details = {}
        values = self.data
        if self.axis_type is AxisType.TIME:
            values = np.array(values).astype(np.datetime64)
            time_unit = time_step = None
            if len(self.data) > 1:
                time_offset = to_offset(pd.Series(values).diff().mode()[0])
                time_unit = time_offset.name
                time_step = time_offset.n
                if time_unit in {
                    "L",
                    "U",
                    "N",
                    "S"
                }:  # skip mili, micro, and nanoseconds
                    values = values.astype(
                        "datetime64[m]"
                    )  # with minute resoluton
                    time_offset = to_offset(
                        pd.Series(values).diff().mode().item()
                    )
                    time_unit = time_offset.name
                    time_step = time_offset.n
                axis_specific_details = {
                    "time_unit": FREQ_CODES[time_unit.upper()],
                    "time_step": time_step,
                }
        elif (
            self.axis_type is AxisType.VERTICAL
            or self.axis_type is AxisType.GENERIC
        ):
            # e.g. numpy.float32 is not JSON serializable
            axis_specific_details = {
                "values": maybe_convert_to_json_serializable(
                    np.unique(np.array(values))
                )
                if unique_values
                else maybe_convert_to_json_serializable(
                    np.array(np.atleast_1d(values))
                )
            }
        return dict(
            **{
                "min": maybe_convert_to_json_serializable(np.nanmin(values)),
                "max": maybe_convert_to_json_serializable(np.nanmax(values)),
                "units": str(self.units),
                "axis": self.axis_type.name,
            },
            **axis_specific_details,
        )

    @classmethod
    @geokube_logging
    def from_xarray(
        cls,
        ds: xr.Dataset,
        ncvar: str,
        id_pattern: Optional[str] = None,
        mapping: Optional[Mapping[str, str]] = None,
        copy: Optional[bool] = False,
    ) -> "Coordinate":
        if not isinstance(ds, xr.Dataset):
            raise TypeError(
                f"Expected type `xarray.Dataset` but provided `{type(ds)}`"
            )

        da = ds[ncvar]
        var = Variable.from_xarray(da, id_pattern=id_pattern, mapping=mapping)
        encoded_ncvar = da.encoding.get("name", ncvar)
        var.encoding.update(name=encoded_ncvar)

        axis_name = Variable._get_name(da, mapping, id_pattern)
        # `axis` attribute cannot be used below, as e.g for EOBS `latitude` has axis `Y`, so wrong AxisType is chosen
        axistype_name = None
        if mapping is not None and da.name in mapping:
            axistype_name = mapping[da.name].get("axis")
        if axistype_name is None:
            axistype_name = da.attrs.get("standard_name", ncvar)
        axistype = AxisType.parse(axistype_name)
        axis = Axis(
            name=axis_name,
            is_dim=ncvar in da.dims,
            axistype=axistype,
            encoding={"name": encoded_ncvar},
        )
        bnds_ncvar = da.encoding.get("bounds", da.attrs.get("bounds"))
        if bnds_ncvar is not None:
            try:
                bnds_name = Variable._get_name(ds[bnds_ncvar], mapping, id_pattern)
                bounds = {
                    bnds_name: Variable.from_xarray(
                        ds[bnds_ncvar],
                        id_pattern=id_pattern,
                        copy=copy,
                        mapping=mapping,
                    )
                }
                if (
                    "units" not in ds[bnds_ncvar].attrs
                    and "units" not in ds[bnds_ncvar].encoding
                ):
                    bounds[bnds_name]._units = var.units
            except:
                bounds = None
        else:
            bounds = None
        return Coordinate(data=var, axis=axis, bounds=bounds)


class ArrayCoordinate(Coordinate):
    pass


class ParametricCoordinate(Coordinate):
    pass



import cf_units as cf
import dask.array as da
import numpy as np
import pytest
import xarray as xr

from geokube.core.axis import Axis, AxisType
from geokube.core.unit import Unit
from geokube.core.variable import Variable
from geokube.utils.attrs_encoding import CFAttributes
from tests.fixtures import *


def test_fails_on_wrong_type():
    with pytest.raises(
        TypeError,
        match=(
            r"Expected argument is one of the following types `number.Number`,"
            r" `numpy.ndarray`, `dask.array.Array`, or `xarray.Variable`*"
        ),
    ):
        _ = Variable({1, 2, 3, 4})

    with pytest.raises(
        TypeError,
        match=(
            r"Expected argument is one of the following types `number.Number`,"
            r" `numpy.ndarray`, `dask.array.Array`, or `xarray.Variable`*"
        ),
    ):
        _ = Variable([1, 2, 3, 4])

    with pytest.raises(
        TypeError,
        match=(
            r"Expected argument is one of the following types `number.Number`,"
            r" `numpy.ndarray`, `dask.array.Array`, or `xarray.Variable`*"
        ),
    ):
        _ = Variable("some_data")


def test_init_proper_attrs_set():
    d = np.random.random((10, 50))
    v = Variable(
        data=d,
        dims=[AxisType.LATITUDE, AxisType.LONGITUDE],
        units="m",
        properties={"prop1": "aaa"},
        encoding={"enc1": "bbb"},
    )

    assert set(v.dim_names) == {"latitude", "longitude"}
    assert set(v.dim_ncvars) == {"latitude", "longitude"}
    assert v.properties == {"prop1": "aaa"}
    assert v.units == Unit("m")


def test_init_based_on_other_variable():
    d = np.random.random((10, 50))
    v = Variable(
        data=d,
        dims=[AxisType.LATITUDE, AxisType.LONGITUDE],
        units="m",
        properties={"prop1": "aaa"},
        encoding={"enc1": "bbb"},
    )
    v2 = Variable(data=v)
    assert id(v) != id(v2)
    assert np.all(v._data == v2._data)
    assert id(v._data) == id(v2._data)
    assert set(v2.dim_names) == {"latitude", "longitude"}
    assert set(v2.dim_ncvars) == {"latitude", "longitude"}
    assert v2.properties == {"prop1": "aaa"}
    assert v2.units == Unit("m")

    xrv = v2.to_xarray(encoding=True)
    assert isinstance(xrv, xr.Variable)
    assert xrv.attrs == dict(**v.properties, units="m")
    assert xrv.encoding == v.encoding
    assert np.all(xrv.values == d)
    assert set(xrv.dims) == {"latitude", "longitude"}

    v3 = Variable.from_xarray(xr.DataArray(xrv))
    assert id(v3) != id(v2)
    assert id(v3._data) == id(v2._data)
    assert set(v3.dim_names) == {"latitude", "longitude"}
    assert set(v3.dim_ncvars) == {"latitude", "longitude"}
    assert v3.properties == {"prop1": "aaa"}
    assert v3.units == Unit("m")


def test_init_proper_attr_set():
    d = np.random.random((10, 50, 20))
    v = Variable(
        data=d,
        dims=[
            Axis("lat", is_dim=True),
            Axis("lon", is_dim=True),
            Axis("depth", is_dim=True),
        ],
        units="m",
        properties={"prop1": "aaa"},
        encoding={"enc1": "bbb"},
    )
    assert set(v.dim_ncvars) == {"lat", "lon", "depth"}
    xrv = v.to_xarray(encoding=True)
    assert isinstance(xrv, xr.Variable)
    assert xrv.attrs == v.properties
    assert xrv.encoding == v.encoding
    assert np.all(xrv.values == d)
    assert set(xrv.dims) == {"lat", "lon", "depth"}

    xrv = v.to_xarray(encoding=False)
    assert isinstance(xrv, xr.Variable)
    assert xrv.attrs == v.properties
    assert xrv.encoding == v.encoding
    assert np.all(xrv.values == d)
    assert set(xrv.dims) == {"lat", "lon", "depth"}


def test_init_proper_attr_set_with_encoding_for_axes():
    d = da.random.random((10, 50, 20))
    v = Variable(
        data=d,
        dims=[
            Axis(
                name="lat",
                axistype=AxisType.LATITUDE,
                is_dim=True,
                encoding={"name": "latT"},
            ),
            Axis(
                name="lon",
                axistype=AxisType.LONGITUDE,
                is_dim=True,
                encoding={"name": "lonN"},
            ),
            Axis(
                name="depth",
                axistype=AxisType.VERTICAL,
                is_dim=True,
                encoding={"name": "depthH"},
            ),
        ],
        units="m",
        properties={"prop1": "aaa"},
        encoding={"enc1": "bbb"},
    )
    assert isinstance(v._data, da.Array)
    assert set(v.dim_ncvars) == {"latT", "lonN", "depthH"}
    assert set(v.dim_names) == {"lat", "lon", "depth"}
    assert v.units == Unit("m")

    xrv = v.to_xarray(encoding=True)
    assert isinstance(xrv, xr.Variable)
    assert xrv.attrs == v.properties
    assert xrv.encoding == v.encoding
    assert np.all(xrv.values == d)
    assert set(xrv.dims) == {"latT", "lonN", "depthH"}

    xrv = v.to_xarray(encoding=False)
    assert isinstance(xrv, xr.Variable)
    assert xrv.attrs == v.properties
    assert xrv.encoding == v.encoding
    assert np.all(xrv.values == d)
    assert set(xrv.dims) == {"lat", "lon", "depth"}
    assert xrv.attrs["units"] == "m"


def test_from_xarray_with_id_pattern(era5_netcdf):
    v = Variable.from_xarray(
        era5_netcdf["tp"], id_pattern="prefix:{units}_{long_name}"
    )

    d1 = v.dims[0]
    assert d1.type is AxisType.TIME
    assert (
        d1.name == "time"
    )  # if any id_pattern component is not found, then defaults is taken

    d2 = v.dims[1]
    assert d2.type is AxisType.LATITUDE
    assert d2.name == "prefix:degrees_north_latitude"
    assert d2.ncvar == "latitude"

    d3 = v.dims[2]
    assert d3.type is AxisType.LONGITUDE
    assert d3.name == "prefix:degrees_east_longitude"
    assert d3.ncvar == "longitude"

    xrv1 = v.to_xarray(encoding=True)
    assert xrv1.attrs == era5_netcdf["tp"].attrs
    assert xrv1.encoding == era5_netcdf["tp"].encoding
    assert set(xrv1.dims) == set(era5_netcdf["tp"].dims)

    xrv1 = v.to_xarray(encoding=False)
    assert xrv1.attrs == era5_netcdf["tp"].attrs
    assert xrv1.encoding == era5_netcdf["tp"].encoding
    assert set(xrv1.dims) == {
        "prefix:degrees_east_longitude",
        "prefix:degrees_north_latitude",
        "time",
    }


def test_from_xarray_rotated_pole_with_mapping(era5_rotated_netcdf_wso):
    v = Variable.from_xarray(
        era5_rotated_netcdf_wso["W_SO"],
        mapping={"soil1": {"name": "my_depth"}},
    )
    assert set(v.dim_ncvars) == (
        set(era5_rotated_netcdf_wso.dims.keys()) - {"bnds"}
    )
    mask = ~np.isnan(v._data)
    assert np.allclose(
        np.array(v._data)[mask],
        np.array(era5_rotated_netcdf_wso["W_SO"]._variable._data[mask]),
    )
    r1 = v.to_xarray(encoding=True)
    assert r1.dims == era5_rotated_netcdf_wso["W_SO"].dims
    assert r1.encoding == era5_rotated_netcdf_wso["W_SO"].encoding
    assert r1.attrs == era5_rotated_netcdf_wso["W_SO"].attrs

    r2 = v.to_xarray(encoding=False)
    assert set(r2.dims) == {
        "my_depth",
        "time",
        "grid_latitude",
        "grid_longitude",
    }
    assert r2.encoding == era5_rotated_netcdf_wso["W_SO"].encoding
    assert r2.attrs == era5_rotated_netcdf_wso["W_SO"].attrs



from __future__ import annotations

from collections.abc import Sequence
import functools as ft
import json
import os
import uuid
import tempfile
import warnings
from html import escape
from itertools import chain
from numbers import Number
from typing import (
    Any,
    Callable,
    Hashable,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)
from warnings import warn

import cartopy.crs as ccrs
import cartopy.feature as cartf
import dask.array as da
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import pyarrow as pa
from shapely.geometry import MultiPolygon, Polygon
from shapely.vectorized import contains as sv_contains
import xarray as xr
import hvplot.xarray  # noqa
from dask import is_dask_collection
from xarray.core.options import OPTIONS
from ..utils import formatting, formatting_html, util_methods
from ..utils.decorators import geokube_logging
from ..utils.util_methods import convert_cftimes_to_numpy
from ..utils.hcube_logger import HCubeLogger
from .axis import Axis, AxisType
from .errs import EmptyDataError
from .cell_methods import CellMethod
from .coord_system import CoordSystem, GeogCS, RegularLatLon, RotatedGeogCS, TransverseMercator, WebMercator
from .coordinate import Coordinate, CoordinateType
from .domain import Domain, DomainType, GeodeticPoints, GeodeticGrid
from .enums import MethodType, RegridMethod
from .unit import Unit
from .variable import Variable
from .domainmixin import DomainMixin

_CARTOPY_FEATURES = {
    "borders": cartf.BORDERS,
    "coastline": cartf.COASTLINE,
    "lakes": cartf.LAKES,
    "land": cartf.LAND,
    "ocean": cartf.OCEAN,
    "rivers": cartf.RIVERS,
    "states": cartf.STATES,
}


# pylint: disable=missing-class-docstring


class Field(Variable, DomainMixin):
    __slots__ = (
        "_name",
        "_domain",
        "_cell_methods",
        "_ancillary_data",
        "_id_pattern",
        "_mapping",
    )

    _LOG = HCubeLogger(name="Field")

    def __init__(
        self,
        data: Union[Number, np.ndarray, da.Array, xr.Variable, Variable],
        name: str,
        dims: Optional[Union[Tuple[Axis], Tuple[AxisType], Tuple[str]]] = None,
        coords: Optional[
            Union[
                Domain,
                Mapping[str, Union[Number, np.ndarray, da.Array]],
                Mapping[
                    str,
                    Tuple[
                        Tuple[str, ...], Union[Number, np.ndarray, da.Array]
                    ],
                ],
            ]
        ] = None,
        crs: Optional[CoordSystem] = None,
        units: Optional[Union[Unit, str]] = None,
        properties: Optional[Mapping[Hashable, str]] = None,
        encoding: Optional[Mapping[Hashable, str]] = None,
        cell_methods: Optional[CellMethod] = None,
        ancillary: Optional[
            Mapping[Hashable, Union[np.ndarray, Variable]]
        ] = None,
    ) -> None:
        super().__init__(
            data=data,
            units=units,
            dims=dims,
            properties=properties,
            encoding=encoding,
        )
        self._ancillary = None
        self._name = name
        self._domain = (
            coords
            if isinstance(coords, Domain)
            else Domain._make_domain_from_coords_dict_dims_and_crs(
                coords=coords, dims=dims, crs=crs
            )
        )

        self._cell_methods = cell_methods
        if ancillary is not None:
            if not isinstance(ancillary, dict):
                raise TypeError(
                    "Expected type of `ancillary` argument is dict, but the"
                    f" provided one if {type(ancillary)}"
                )
            res_anc = {}
            for k, v in ancillary.items():
                if not isinstance(v, (np.ndarray, da.Array, Variable, Number)):
                    raise TypeError(
                        "Expected type of single ancillary variable is:"
                        " `numpy.ndarray`, `dask.Array`, `geokube.Variable`,"
                        f" or `Number`, but the provided one if {type(v)}"
                    )
                # TODO: what should be axis and dims for ancillary variables? SHould it be `Variable`?
                res_anc[k] = Variable(data=v)
            self._ancillary = res_anc

    def __str__(self) -> str:
        return (
            f"Field {self.name}:{self.ncvar} with cell method:"
            f" {self.cell_methods}"
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def ncvar(self) -> str:
        return self._encoding.get("name", self.name)

    @property
    def cell_methods(self) -> Optional[CellMethod]:
        return self._cell_methods

    @property
    def domain(self) -> Domain:
        return self._domain

    @property
    def coords(self):
        return self._domain._coords

    @property
    def ancillary(self) -> Optional[Mapping[Hashable, Variable]]:
        return self._ancillary

    def __contains__(self, key):
        return key in self.domain

    def __getitem__(self, key):
        return self.domain[key]

    def __repr__(self) -> str:
        return self.to_xarray(encoding=False).__repr__()

    #        return formatting.array_repr(self.to_xarray())

    def _repr_html_(self):
        return self.to_xarray(encoding=False)._repr_html_()
        # if OPTIONS["display_style"] == "text":
        #     return f"<pre>{escape(repr(self.to_xarray()))}</pre>"
        # return formatting_html.array_repr(self)

    def __next__(self):
        for k, v in self.domain._coords.items():
            yield k, v
        raise StopIteration

    # geobbox and locations operates also on dependent coordinates
    # they refer only to GeoCoordinates (lat/lon)
    # TODO: Add Vertical
    @geokube_logging
    def geobbox(
        self,
        north: Number | None = None,
        south: Number | None = None,
        west: Number | None = None,
        east: Number | None = None,
        top: Number | None = None,
        bottom: Number | None = None,
    ) -> Field:
        """
        Subset a field using a bounding box.

        Subsets the original field with the given bounding box.  If a
        bound is omitted or `None`, no subsetting takes place in that
        direction.  At least one bound must be provided.

        Parameters
        ----------
        north, south, west, east : number or None, optional
            Horizontal bounds.
        top, bottom : number or None, optional
            Vertical bounds.

        Returns
        -------
        Field
            A field with the coordinate values between given bounds.

        Raises
        ------
        KeyError
            If no bound is provided.

        """
        if not util_methods.is_atleast_one_not_none(
            north, south, west, east, top, bottom
        ):
            raise KeyError(
                "At least on of the following must be defined: [north, south,"
                " west, east, top, bottom]!"
            )
        return self._geobbox_idx(
            south=south,
            north=north,
            west=west,
            east=east,
            top=top,
            bottom=bottom,
        )

    def _geobbox_cartopy(self, south, north, west, east, top, bottom):
        # TODO: add vertical also
        domain = self._domain

        ind_lat = domain.is_latitude_independent
        ind_lon = domain.is_longitude_independent
        if ind_lat and ind_lon:
            idx = {
                domain.latitude.name: np.s_[south:north]
                if util_methods.is_nondecreasing(domain.latitude.data)
                else np.s_[north:south],
                domain.longitude.name: np.s_[west:east]
                if util_methods.is_nondecreasing(domain.longitude.data)
                else np.s_[east:west],
            }
            return self.sel(indexers=idx, roll_if_needed=roll_if_needed)

        # Specifying the corner points of the bounding box in the rectangular
        # coordinate system (`cartopy.crs.PlateCarree()`).
        lats = np.array([south, south, north, north], dtype=np.float32)
        lons = np.array([west, east, west, east], dtype=np.float32)

        # Transforming the corner points of the bounding box from the
        # rectangular coordinate system (`cartopy.crs.PlateCarree`) to the
        # coordinate system of the field.
        plate = ccrs.PlateCarree()
        pts = domain.crs.as_cartopy_crs().transform_points(
            src_crs=plate, x=lons, y=lats
        )
        x, y = pts[:, 0], pts[:, 1]

        # Spatial subseting.
        idx = {
            domain[AxisType.LATITUDE].dims[1].ncvar: np.s_[x.min() : x.max()],
            domain[AxisType.LATITUDE].dims[0].ncvar: np.s_[y.min() : y.max()],
        }
        ds = (
            self._check_and_roll_longitude(self.to_xarray(), idx)
            if roll_if_needed
            else self.to_xarray()
        )
        return Field.from_xarray(
            ds=ds.sel(indexers=idx),
            ncvar=self.ncvar,
            copy=False,
            id_pattern=self._id_pattern,
            mapping=self._mapping,
        )

    def _geobbox_idx(
        self,
        south: Number,
        north: Number,
        west: Number,
        east: Number,
        top: Number | None = None,
        bottom: Number | None = None,
    ):
        field = self
        lat = field.domain.get(AxisType.LATITUDE)
        lon = field.domain.get(AxisType.LONGITUDE)

#        lat, lon = field.latitude, field.longitude

        # Vertical
        # NOTE: In this implementation, vertical is always considered an
        # independent coordinate.
        if top is not None or bottom is not None:
            try:
                vert = field.vertical
            except KeyError:
                vert = None
            # TODO: Reconsider `not vert.shape`.
            if vert is None or not vert.shape:
                raise ValueError(
                    "'top' and 'bottom' must be None because there is no "
                    "vertical coordinate or it is constant"
                )
            vert_incr = util_methods.is_nondecreasing(vert.data)
            if vert.attrs.get("positive") == "down":
                top = None if top is None else -top
                bottom = None if bottom is None else -bottom
                vert_incr = ~vert_incr
            vert_slice = np.s_[bottom:top] if vert_incr else np.s_[top:bottom]
            vert_idx = {vert.name: vert_slice}
            field = field.sel(indexers=vert_idx, roll_if_needed=True)

        idx = {}
        independent = False
        if lat is not None and field.is_latitude_independent:
            # Case of latitude and longitude being independent.
            lat_incr = util_methods.is_nondecreasing(lat.data)
            lat_slice = np.s_[south:north] if lat_incr else np.s_[north:south]
            idx[lat.name] = lat_slice
            independent = True

        if lon is not None and field.is_longitude_independent:
            lon_incr = util_methods.is_nondecreasing(lon.data)
            lon_slice = np.s_[west:east] if lon_incr else np.s_[east:west]
            idx[lon.name] = lon_slice
            independent = True

        if independent:
            return field.sel(indexers=idx, roll_if_needed=True)

        if lat is not None and (not field.is_latitude_independent):
            # Case of latitude and longitude being dependent.
            # Specifying the mask(s) and extracting the indices that correspond
            # to the inside the bounding box.
            lat_mask = util_methods.is_between(lat.data, south, north)
            # TODO: Clarify why this is required.
            if lat_mask.sum() == 0:
                lat_mask = util_methods.is_between(lat.data, north, south)
        else:
            lat_mask = None

        if lon is not None and (not field.is_longitude_independent):
            lon_mask = util_methods.is_between(lon.data, west, east)
            if lon_mask.sum() == 0:
                lon_mask = util_methods.is_between(lon.data, east, west)
        else:
            lon_mask = None

        if (lat_mask is not None and lon_mask is not None):
            nonzero_idx = np.nonzero(lat_mask & lon_mask)
            idx = {
                lat.dims[i].name: np.s_[incl_idx.min() : incl_idx.max() + 1]
                for i, incl_idx in enumerate(nonzero_idx)
            }
        elif lat_mask is not None:
            nonzero_idx = np.nonzero(lat_mask)
            idx = idx.update({
                lat.dims[i].name: np.s_[incl_idx.min() : incl_idx.max() + 1]
                for i, incl_idx in enumerate(nonzero_idx)
            })
        elif lon_mask is not None:
            nonzero_idx = np.nonzero(lon_mask)
            idx = idx.update({
                lon.dims[i].name: np.s_[incl_idx.min() : incl_idx.max() + 1]
                for i, incl_idx in enumerate(nonzero_idx)
            })
        else:
            warn("'field' does not have latitude nor longitude dimensions")

        dset = field.to_xarray(encoding=False)
        dset = field._check_and_roll_longitude(dset, idx)
        dset = dset.isel(indexers=idx)

        return Field.from_xarray(
            ds=dset,
            ncvar=self.name,
            copy=False,
            id_pattern=self._id_pattern,
            mapping=self._mapping,
        )

    def locations(
        self,
        latitude: Number | Sequence[Number],
        longitude: Number | Sequence[Number],
        vertical: Number | Sequence[Number] | None = None,
    ) -> (
        Field
    ):  # points are expressed as arrays for coordinates (dep or ind) lat/lon/vertical
        """
        Select points with given coordinates from a field.

        Subsets the original field by selecting only the points with
        provided coordinates and returns a new field with these points.
        Uses the nearest neighbor method.  The resulting field has a
        domain with the points nearest to the provided coordinates.

        Parameters
        ----------
        latitude, longitude : array-like or number
            Latitude and longitude coordinate values.  Must be of the
            same shape.
        vertical : array-like or number or None, optional
            Verical coordinate values.  If given and not `None`, must be
            of the same shape as `latitude` and `longitude`.

        Returns
        -------
        Field
            A field with a point domain that contains given locations.

        Examples
        --------
        >>> result = field.locations(latitude=40, longitude=35)
        >>> result.latitude.values
        array([40.86], dtype=float32)
        >>> result.longitude.values
        array([34.99963], dtype=float32)

        Vertical coordinate is optional.  If provided, the vertical axis
        of the resulting field is also expressed with points:

        >>> result = field.locations(
        ...     latitude=40,
        ...     longitude=35,
        ...     vertical=-2
        ... )
        >>> result.latitude.values
        array([40.86], dtype=float32)
        >>> result.longitude.values
        array([34.99963], dtype=float32)
        >>> result.vertical.values
        array([2.5010786], dtype=float32)

        It is possible to provide the coordinates of multiple points at
        once with an array-like object.  In that case, `latitude`,
        `longitude`, and `vertical` must have the same length.

        >>> result = temperature_field.locations(
        ...     latitude=[40, 41],
        ...     longitude=[32, 35],
        ...     vertical=[-2, -5]
        ... )
        >>> result.latitude.values
        array([40.86   , 40.99889], dtype=float32)
        >>> result.longitude.values
        array([31.99963, 34.99963], dtype=float32)
        >>> result.vertical.values
        array([2.5010786, 2.5010786], dtype=float32)
        """
        return self._locations_idx(
            latitude=latitude, longitude=longitude, vertical=vertical
        )

    def interpolate(self, domain: Domain, method: str = "nearest") -> Field:
        # TODO: Add vertical support.
        # if (
        #     {c.axis_type for c in domain.coords.values() if c.is_dimension}
        #     != {AxisType.LATITUDE, AxisType.LONGITUDE}
        # ):
        #     raise NotImplementedError(
        #         "'domain' can have only latitude and longitude at the moment"
        #     )

        dset = self.to_xarray(encoding=False)
        lat, lon = domain.latitude.values, domain.longitude.values
        if self.is_latitude_independent and self.is_longitude_independent:
            if domain.type is DomainType.POINTS:
                dim_lat = dim_lon = "points"
            else:
                dim_lat, dim_lon = self.latitude.name, self.longitude.name
            interp_coords = {
                self.latitude.name: xr.DataArray(data=lat, dims=dim_lat),
                self.longitude.name: xr.DataArray(data=lon, dims=dim_lon),
            }
            dset_interp = dset.interp(coords=interp_coords, method=method)
        else:
            if domain.type is DomainType.POINTS:
                pts = self.domain.crs.as_cartopy_crs().transform_points(
                    src_crs=domain.crs.as_cartopy_crs(), x=lon, y=lat
                )
                x, y = pts[..., 0], pts[..., 1]
                interp_coords = {
                    self.x.name: xr.DataArray(data=x, dims="points"),
                    self.y.name: xr.DataArray(data=y, dims="points"),
                }
            else:
                if lat.ndim == lon.ndim == 1:
                    lon, lat = np.meshgrid(lon, lat)
                pts = self.domain.crs.as_cartopy_crs().transform_points(
                    src_crs=ccrs.PlateCarree(), x=lon, y=lat
                )
                x, y = pts[..., 0], pts[..., 1]
                dims = (domain.latitude.name, domain.longitude.name)
                grid = xr.Dataset(
                    data_vars={self.x.name: (dims, x), self.y.name: (dims, y)},
                    coords=domain.to_xarray(encoding=False),
                )
                dset = dset.drop(
                    labels=(self.latitude.name, self.longitude.name)
                )
                interp_coords = {
                    self.x.name: grid[self.x.name],
                    self.y.name: grid[self.y.name],
                }
            dset_interp = dset.interp(coords=interp_coords, method=method)
            dset_interp = dset_interp.drop(labels=[self.x.name, self.y.name])

        dset_interp[self.name].encoding.update(dset[self.name].encoding)

        encoding_coords = " ".join(
            coord_name
            for coord_name, coord in domain.coords.items()
            if not coord.is_dimension
        )
        # NOTE: we need to manually append scalar coordinates
        # NOTE: as they are not affected by interpolation
        encoding_coords = " ".join(
            chain(
                [encoding_coords],
                [
                    coord_name
                    for coord_name, coord in self.domain.coords.items()
                    if coord.type is CoordinateType.SCALAR
                ],
            )
        )
        if encoding_coords:
            dset_interp[self.name].encoding["coordinates"] = encoding_coords
        # TODO: Fill value should depend on the data type.
        # TODO: Add xarray fillna into Field.to_xarray.
        dset_interp[self.name].encoding["_FillValue"] = -9.0e-20

        field = Field.from_xarray(
            ds=dset_interp,
            ncvar=self.name,
            copy=False,
            id_pattern=self._id_pattern,
            mapping=self._mapping,
        )

        field.domain.type = domain.type
        field.domain.crs = domain.crs
        return field

    def _locations_cartopy(self, latitude, longitude, vertical=None):
        domain = self._domain

        # Specifying the location points in the rectangular coordinate system
        # (`cartopy.crs.PlateCarree()`).
        lats = np.array(latitude, dtype=np.float32, ndmin=1)
        lons = np.array(longitude, dtype=np.float32, ndmin=1)

        ind_lat = domain.is_latitude_independent
        ind_lon = domain.is_longitude_independent
        if ind_lat and ind_lon:
            idx = {
                domain.latitude.name: lats.item() if len(lats) == 1 else lats,
                domain.longitude.name: lons.item() if len(lons) == 1 else lons,
            }
        else:
            # Transforming the location points from the rectangular coordinate
            # system (`cartopy.crs.PlateCarree`) to the coordinate system of
            # the field.
            plate = ccrs.PlateCarree()
            pts = domain.crs.as_cartopy_crs().transform_points(
                src_crs=plate, x=lons, y=lats
            )
            idx = {
                domain.x.name: xr.DataArray(data=pts[:, 0], dims="points"),
                domain.y.name: xr.DataArray(data=pts[:, 1], dims="points"),
            }

        return Field.from_xarray(
            ds=self.to_xarray(encoding=False).sel(
                indexers=idx, method="nearest"
            ),
            ncvar=self.name,
            copy=False,
            id_pattern=self._id_pattern,
            mapping=self._mapping,
        )

    def _locations_idx(self, latitude, longitude, vertical=None):
        field = self
        sel_kwa = {"roll_if_needed": True, "method": "nearest"}
        lats = np.array(latitude, dtype=np.float32).reshape(-1)
        lons = np.array(longitude, dtype=np.float32).reshape(-1)

        n = lats.size
        if lons.size != n:
            raise ValueError(
                "'latitude' and 'longitude' must have the same number of items"
            )

        # Vertical
        # NOTE: In this implementation, vertical is always considered an
        # independent coordinate.
        if vertical is not None:
            verts = np.array(vertical, dtype=np.float32).reshape(-1)
            if verts.size != n:
                raise ValueError(
                    "'vertical' must have the same number of items as "
                    "'latitude' and 'longitude'"
                )
            if field.vertical.attrs.get("positive") == "down":
                verts = -verts
            verts = xr.DataArray(data=verts, dims="points")
            vert_ax = Axis(name=self.vertical.name, axistype=AxisType.VERTICAL)
            field = field.sel(indexers={vert_ax: verts}, **sel_kwa)

        # Case of latitude and longitude being independent.
        if self.is_latitude_independent and self.is_longitude_independent:
            # TODO: Check lon values conventions.
            lats = xr.DataArray(data=lats, dims="points")
            lons = xr.DataArray(data=lons, dims="points")
            idx = {self.latitude.name: lats, self.longitude.name: lons}
            result_field = field.sel(indexers=idx, **sel_kwa)
        else:
            # TODO: Check lon values conventions if possible, otherwise raise error.
            # Case of latitude and longitude being dependent on y and x.
            # Adjusting the shape of the latitude and longitude coordinates.
            # TODO: Check if these are NumPy arrays.
            # TODO: Check axes and shapes manipulation again.
            lat_data = self.latitude.values
            lat_dims = (np.s_[:],) + (np.newaxis,) * lat_data.ndim
            lat_data = lat_data[np.newaxis, :]
            lon_data = self.longitude.values
            lon_dims = (np.s_[:],) + (np.newaxis,) * lon_data.ndim
            lon_data = lon_data[np.newaxis, :]

            # Adjusting the shape of the latitude and longitude of the
            # locations.
            lats = lats[lat_dims]
            lons = lons[lon_dims]

            # Calculating the squares of the Euclidean distance.
            lat_diff = lat_data - lats
            lon_diff = lon_data - lons
            diff_sq = lat_diff * lat_diff + lon_diff * lon_diff

            # Selecting the indices that correspond to the squares of the
            # Euclidean distance.
            # TODO: Improve vectorization.
            # TODO: Consider replacing `numpy.unravel_index` with
            # `numpy.argwhere`, using the constructs like
            # `np.argwhere(diff_sq[i] == diff_sq[i].min())[0]`.
            n, *shape = diff_sq.shape
            idx_ = tuple(
                np.unravel_index(indices=diff_sq[i].argmin(), shape=shape)
                for i in range(n)
            )
            idx_ = np.array(idx_, dtype=np.int64)

            # Spatial subseting.
            idx = {
                dim.name: xr.DataArray(data=idx_[:, i], dims="points")
                for (i,), dim in np.ndenumerate(self.latitude.dims)
            }

            result_dset = field.to_xarray(encoding=False)
            result_dset = field._check_and_roll_longitude(result_dset, idx)
            result_dset = result_dset.isel(indexers=idx)
            result_field = Field.from_xarray(
                ds=result_dset,
                ncvar=self.name,
                copy=False,
                id_pattern=self._id_pattern,
                mapping=self._mapping,
            )

        result_field.domain.crs = RegularLatLon()
        result_field.domain._type = DomainType.POINTS

        return result_field

    # consider only independent coordinates
    # TODO: we should use metpy approach (user can also specify - units)
    @geokube_logging
    def sel(
        self,
        indexers: Mapping[Union[Axis, str], Any] = None,
        roll_if_needed: bool = True,
        method: str = None,
        tolerance: Number = None,
        drop: bool = False,  # TODO: check if should be always True or False in out case
        **indexers_kwargs: Any,
    ) -> "Field":
        indexers = xr.core.utils.either_dict_or_kwargs(
            indexers, indexers_kwargs, "sel"
        )
        # TODO:
        indexers = self.domain.map_indexers(indexers)
        # TODO: indexers from this place should be always of type -> Mapping[Axis, Any]
        #   ^ it can be done in Domain
        indexers = indexers.copy()
        ds = self.to_xarray(encoding=False)

        if (
            time_ind := indexers.get(Axis("time"))
        ) is not None and util_methods.is_time_combo(time_ind):
            # NOTE: time is always independent coordinate
            try:
                idx = self.domain._process_time_combo(time_ind)
            except KeyError:
                self._LOG.warn("time axis is not present in the domain.")
            else:
                if (
                    isinstance(idx["time"], np.ndarray)
                    and len(idx["time"]) == 0
                ):
                    Field._LOG.warn("empty `time` indexer")
                    raise EmptyDataError("empty `time` indexer")
                ds = ds.isel(idx, drop=drop)
                del indexers[Axis("time")]

        if roll_if_needed:
            ds = self._check_and_roll_longitude(ds, indexers)

        indexers = {
            self.domain[k].name: v
            for k, v in indexers.items()
            if k in self.domain
        }
        #indexers = {
        #    index_key: index_value
        #    for index_key, index_value in indexers.items()
        #    if index_key in ds.xindexes
        #}

        # If selection by single lat/lon, coordinate is lost as it is not stored either in da.dims nor in da.attrs["coordinates"]
        # and then selecting this location from Domain fails
        ds_dims = set(ds.dims)
        ds = ds.sel(indexers, tolerance=tolerance, method=method, drop=drop)
        lost_dims = ds_dims - set(ds.dims)
        Field._update_coordinates(ds[self.name], lost_dims)
        return Field.from_xarray(
            ds,
            ncvar=self.name,
            id_pattern=self._id_pattern,
            mapping=self._mapping,
        )

    @geokube_logging
    def _check_and_roll_longitude(self, ds, indexers) -> xr.Dataset:
        # `ds` here is passed as an argument to avoid one redundent to_xarray call
        if Axis("longitude") not in indexers:
            return ds
        if (
            self.domain[Axis("longitude")].type
            is not CoordinateType.INDEPENDENT
        ):
            # TODO: implement for dependent coordinate
            raise NotImplementedError(
                "Rolling longitude is currently supported only for independent"
                " coordinate!"
            )
        first_el, last_el = (
            self.domain[Axis("longitude")].min(),
            self.domain[Axis("longitude")].max(),
        )

        if isinstance(indexers[Axis("longitude")], slice):
            start = indexers[Axis("longitude")].start
            stop = indexers[Axis("longitude")].stop
            start = 0 if start is None else start
            stop = 0 if stop is None else stop
            sel_neg_conv = (start < 0) | (stop < 0)
            sel_pos_conv = (start > 180) | (stop > 180)
        else:
            vals = np.array(indexers[Axis("longitude")], ndmin=1)
            sel_neg_conv = np.any(vals < 0)
            sel_pos_conv = np.any(vals > 180)

        dset_neg_conv = first_el < 0
        dset_pos_conv = first_el >= 0
        lng_name = self.domain[Axis("longitude")].name

        if dset_pos_conv and sel_neg_conv:
            # from [0,360] to [-180,180]
            # Attributes are lost while doing `assign_coords`. They need to be reassigned (e.q. by `update`)
            roll_value = (ds[lng_name] >= 180).sum().item()
            res = ds.assign_coords(
                {lng_name: (((ds[lng_name] + 180) % 360) - 180)}
            ).roll(**{lng_name: roll_value}, roll_coords=True)
            res[lng_name].attrs.update(ds[lng_name].attrs)
            # TODO: verify of there are some attrs that need to be updated (e.g. min/max value)
            return res
        if dset_neg_conv and sel_pos_conv:
            # from [-180,-180] to [0,360]
            roll_value = (ds[lng_name] <= 0).sum().item()
            res = (
                ds.assign_coords({lng_name: (ds[lng_name] % 360)})
                .roll(**{lng_name: -roll_value}, roll_coords=True)
                .assign_attrs(**ds[lng_name].attrs)
            )
            res[lng_name].attrs.update(ds[lng_name].attrs)
            return res
        return ds

    def to_regular(self):
        # Infering latitude and longitude steps from the x and y coordinates.
        if isinstance(self.domain.crs, RotatedGeogCS):
            lat_step = np.ptp(self.y.values) / (self.y.values.size - 1)
            lon_step = np.ptp(self.x.values) / (self.x.values.size - 1)
        else:
            raise NotImplementedError(
                f"'{type(self.domain.crs).__name__}' is not supported as a "
                "type of coordinate reference system"
            )

        # Building regular latitude-longitude coordinates.
        south = self.latitude.values.min()
        north = self.latitude.values.max()
        west = self.longitude.values.min()
        east = self.longitude.values.max()
        lat = np.arange(south, north + lat_step / 2, lat_step)
        lon = np.arange(west, east + lon_step / 2, lon_step)

        return self.interpolate(
            domain=GeodeticGrid(latitude=lat, longitude=lon), method="nearest"
        )

    def extract_polygons(self, geometry, crop=True, return_mask=False):
        # Preparing geometry.
        polygons = []
        for polygon in np.asarray(geometry).flat:
            if isinstance(polygon, Polygon):
                polygons.append(polygon)
            elif isinstance(polygon, MultiPolygon):
                polygons += list(polygon.geoms)
            else:
                raise TypeError(
                    "'geometry' must contain one or more instances of "
                    f"'Polygon' or'MultiPolygon', not {type(polygon)}"
                )
        multi_polygon = MultiPolygon(polygons=polygons)

        # Preparing masks.
        # HACK: Check against `None` is provided temporarily.
        if (
            self.domain.type is not DomainType.GRIDDED
            and self.domain.type is not None
        ):
            raise NotImplementedError(
                "'self.domain.type' must be 'DomainType.GRIDDED'"
            )
        field = (
            self if isinstance(self.domain.crs, GeogCS) else self.to_regular()
        )
        lat, lon = field.latitude.values, field.longitude.values
        coords = {"latitude": lat, "longitude": lon}
        mask = xr.DataArray(coords=coords, dims=coords.keys(), name="mask")
        lon_, lat_ = np.meshgrid(lon, lat, indexing="xy")
        mask.values = sv_contains(geometry=multi_polygon, x=lon_, y=lat_)

        # Applying mask.
        data = field.to_xarray(encoding=False).where(mask, other=np.nan)

        # Cropping.
        if crop:
            lon_min, lat_min, lon_max, lat_max = multi_polygon.bounds
            lat_order = 1 if lat[0] <= lat[-1] else -1
            lon_order = 1 if lon[0] <= lon[-1] else -1
            idx = {
                field.latitude.name: np.s_[lat_min:lat_max:lat_order],
                field.longitude.name: np.s_[lon_min:lon_max:lon_order],
            }
            data = data.sel(indexers=idx)

        # Converting back to field.
        result = Field.from_xarray(data, ncvar=field.name)

        return (result, mask) if return_mask else result

    # TO CHECK
    @geokube_logging
    def regrid(
        self,
        target: Union[Domain, "Field"],
        method: Union[str, RegridMethod] = "bilinear",
        weights_path: Optional[str] = None,
        reuse_weights: bool = True,
    ) -> "Field":
        """
        Regridds present coordinate system.
        Parameters
        ----------
        target_domain : geokube.Domain or geokube.Field
            Domain which is supposed to be the result of regridding.
        method : str
            A method to use for regridding. Default: `bilinear`.
        weights_path : str, optional
            The path of the file where the interpolation weights are
            stored. Default: `None`.
        reuse_weights : bool, optional
            Whether to reuse already calculated weights or not. Default:
            `True`.
        Returns
        ----------
        field : Field
           The field with values modified by regridding query.
        Examples:
        ----------
        >>> result = field.regrid(
        ...     target_domain=target_domain,
        ...     method='bilinear'
        ... )

        """
        import xesmf as xe
        if not isinstance(target, Domain):
            if isinstance(target, Field):
                target = target.domain
            else:
                raise TypeError(
                    "'target' must be an instance of Domain or Field"
                )

        if not isinstance(method, RegridMethod):
            method = RegridMethod[str(method).upper()]

        if reuse_weights and (
            weights_path is None or not os.path.exists(weights_path)
        ):
            Field._LOG.warn("`weights_path` is None or file does not exist!")
            Field._LOG.info("`reuse_weights` turned off")
            reuse_weights = False

        names_in = {self.latitude.name: "lat", self.longitude.name: "lon"}
        names_out = {target.latitude.name: "lat", target.longitude.name: "lon"}
        coords_in = coords_out = None

        if method in {
            RegridMethod.CONSERVATIVE,
            RegridMethod.CONSERVATIVE_NORMED,
        }:
            self.domain.compute_bounds()
            lat_b_name = next(iter(self.latitude.bounds))
            lat_b = next(iter(self.latitude.bounds.values())).values
            lat_b = Domain.convert_bounds_2d_to_1d(lat_b)
            lon_b_name = next(iter(self.longitude.bounds))
            lon_b = next(iter(self.longitude.bounds.values())).values
            lon_b = Domain.convert_bounds_2d_to_1d(lon_b)
            names_in.update({lat_b_name: "lat_b", lon_b_name: "lon_b"})
            coords_in = {"lat_b": lat_b, "lon_b": lon_b}

            target.compute_bounds()
            lat_b_name = next(iter(target.latitude.bounds))
            lat_b = next(iter(target.latitude.bounds.values())).values
            lat_b = Domain.convert_bounds_2d_to_1d(lat_b)
            lon_b_name = next(iter(target.longitude.bounds))
            lon_b = next(iter(target.longitude.bounds.values())).values
            lon_b = Domain.convert_bounds_2d_to_1d(lon_b)
            names_out.update({lat_b_name: "lat_b", lon_b_name: "lon_b"})
            coords_out = {"lat_b": lat_b, "lon_b": lon_b}

        # Regridding
        in_ = self.to_xarray(encoding=False).rename(names_in)
        if coords_in:
            in_ = in_.assign_coords(coords=coords_in)
        out = target.to_xarray(encoding=False).to_dataset().rename(names_out)
        if coords_out:
            out = out.assign_coords(coords=coords_out)
        regrid_kwa = {
            "ds_in": in_,
            "ds_out": out,
            "method": method.value,
            "unmapped_to_nan": True,
            "filename": weights_path,
        }
        try:
            regridder = xe.Regridder(**regrid_kwa, reuse_weights=reuse_weights)
        except PermissionError:
            regridder = xe.Regridder(**regrid_kwa)
        result = regridder(in_, keep_attrs=True, skipna=False)
        result = result.rename({v: k for k, v in names_in.items()})
        result[self.name].encoding = in_[self.name].encoding
        if not isinstance(target.crs, GeogCS):
            missing_coords = {
                coord.name: out.coords[coord.name]
                for coord in (target.x, target.y)
                if coord.name not in result.coords
            }
            result = result.assign_coords(coords=missing_coords)
            result[self.name].encoding["coordinates"] = "latitude longitude"
        # After regridding those attributes are not valid!
        util_methods.clear_attributes(result, attrs="cell_measures")
        field_out = Field.from_xarray(
            ds=result,
            ncvar=self.name,
            copy=False,
            id_pattern=self._id_pattern,
            mapping=self._mapping,
        )
        field_out.domain._crs = target.crs
        field_out.domain._type = target.type
        return field_out

    # TO CHECK
    @geokube_logging
    def resample(
        self,
        operator: Union[Callable, MethodType, str],
        frequency: str,
        **resample_kwargs,
    ) -> "Field":
        """
        Perform resampling along the available `time` coordinate.
        Adjust appropriately time bounds.

        Parameters
        ----------
        operator : callable or str
            Callable-object used for aggregation or string
            representation of a function.  Currently supported are the
            methods of ``geokube.MethodType``.
        frequency :  str
            Expected resampling frequency.

        Returns
        ----------
        field : Field
            The field with values after resampling procedure.

        Examples:
        ----------
        Resample to day frequency taking the maximum over the elements
        in each day:
        >>> res = field.resample("maximum", frequency='1D')

        Resample to two-monts frequency taking the sum over each two
        months:
        >>> resulting_field = field.resample("sum", frequency='2M')

        """
        if callable(operator):
            func = operator
        else:
            if isinstance(operator, str):
                operator_func = MethodType(operator)
            elif isinstance(operator, MethodType):
                operator_func = operator
            else:
                raise TypeError(
                    "Operator must be `str`, `MethodType`, or `callable`."
                    " Provided `{type(operator)}`"
                )
            if operator_func is MethodType.UNDEFINED:
                methods = {
                    method.value[0]
                    for method in MethodType.__members__.values()
                }
                methods.discard("<undefined>")
                raise ValueError(
                    f"Provided operator '{operator}' was not found! Available"
                    f" operators are: {sorted(methods)}!"
                )
            func = (
                operator_func.dask_operator
                if is_dask_collection(self)
                else operator_func.numpy_operator
            )

        # ################## Temporary solution for time bounds adjustmnent ######################

        # TODO: handle `formula_terms` bounds attribute
        # http://cfconventions.org/cf-conventions/cf-conventions.html#cell-boundaries
        ds = self.to_xarray(encoding=False)
        if (time_bnds := self.time.bounds) is None:
            bnds_name, bnds = f"{self.time.name}_bnds", None
        else:
            ((bnds_name, bnds),) = time_bnds.items()
        if self.cell_methods and bnds is not None:
            # `closed=right` set by default for {"M", "A", "Q", "BM", "BA", "BQ", "W"} resampling codes ("D" not included!)
            # https://github.com/pandas-dev/pandas/blob/7c48ff4409c622c582c56a5702373f726de08e96/pandas/core/resample.py#L1383
            resample_kwargs.update({"closed": "right"})
            da = ds.resample(
                indexer={self.time.name: frequency}, **resample_kwargs
            )
            new_bnds = np.empty(
                shape=(len(da.groups), 2), dtype=np.dtype("datetime64[ns]")
            )
            for i, v in enumerate(da.groups.values()):
                new_bnds[i] = [bnds.values[v].min(), bnds.values[v].max()]
            # TODO: Check if this is redundant
            if bnds is None:
                Field._LOG.warn(
                    "Time bounds not defined for the cell methods!"
                )
                warnings.warn("Time bounds not defined for the cell methods!")
        else:
            da = ds.resample(
                indexer={self.time.name: frequency}, **resample_kwargs
            )
            new_bnds = np.empty(
                shape=(len(da.groups), 2), dtype=np.dtype("datetime64[ns]")
            )
            for i, v in enumerate(da.groups.values()):
                new_bnds[i] = [
                    self.time.values[v].min(),
                    self.time.values[v].max(),
                ]

        # NOTE: `reduce` spans result for all intermediate values
        # if 1st and 3rd day os selected and 1D freq is used
        # then `reduce` results in time axis for 1st, 2nd, and 3rd
        # passing for missing dates `NaN`s
        da = da.reduce(func=func, dim=self.time.name, keep_attrs=True).dropna(
            dim=self.time.name, how="all"
        )
        # NOTE: `reduce` removes all `encoding` properties
        da[self.name].encoding = ds[self.name].encoding
        res = xr.Dataset(
            da,
            coords={f"{bnds_name}": ((self.time.name, "bnds"), new_bnds)},
        )
        field = Field.from_xarray(res, ncvar=self.name)
        field.time.bounds = new_bnds
        field.domain.crs = self.domain.crs
        field.domain._type = self.domain._type

        # TODO: adjust cell_methods after resampling!

        # #########################################################################################
        return field

    @geokube_logging
    def average(self, dim: str | None = None) -> Field:
        dset = self.to_xarray(encoding=False)
        if dim is None:
            # return dset[self._name].mean().data
            result = dset[self._name].mean().data
            result[self._name].encoding = dset[self._name].encoding
            return Field.from_xarray(
                ds=xr.DataArray(data=result).to_dataset(name=self._name),
                ncvar=self.name,
                id_pattern=self._id_pattern,
                mapping=self._mapping,
                copy=False,
            )
        if self.coords[dim].is_dim:
            result_dset = dset.mean(dim=dim)
            result_dset[self._name].encoding = dset[self._name].encoding
            result_field = Field.from_xarray(
                ds=result_dset,
                ncvar=self.name,
                copy=False,
                id_pattern=self._id_pattern,
                mapping=self._mapping,
            )
            result_field.domain.crs = self._domain.crs
            result_field.domain._type = self._domain._type
            return result_field
        raise ValueError(f"'dim' {dim} is not supported for averaging")

    @geokube_logging
    def to_netcdf(self, path):
        self.to_xarray().to_netcdf(path=path)

    @geokube_logging
    def to_netcdf(self, path):
        self.to_xarray().to_netcdf(path=path)

    # TO CHECK
    @geokube_logging
    def plot(
        self,
        features=None,
        gridlines=None,
        gridline_labels=None,
        subplot_kwargs=None,
        projection=None,
        figsize=None,
        robust=None,
        aspect=None,
        save_path=None,
        save_kwargs=None,
        clean_image=False,
        vmin=None,
        vmax=None,
        normalize=False,
        **kwargs,
    ):
        axis_names = self.domain._axis_to_name
        time = self.coords.get(axis_names.get(AxisType.TIME))
        vert = self.coords.get(axis_names.get(AxisType.VERTICAL))
        lat = self.coords.get(axis_names.get(AxisType.LATITUDE))
        lon = self.coords.get(axis_names.get(AxisType.LONGITUDE))

        # NOTE: The argument `save_kwargs` passes the keyword arguments to the
        # `savefig` method, see:
        # https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.savefig.html

        # Resolving time series and layers because they do not require most of
        # processing other plot types do:
        if self._domain._type is DomainType.POINTS:
            if figsize is not None:
                kwargs["figsize"] = figsize
            n_pts = lat.size
            n_time = time.size if (time is not None and time.is_dim) else 0
            n_vert = vert.size if (vert is not None and vert.is_dim) else 0
            with np.nditer((lat.values, lon.values)) as it:
                # points = [
                #     f"{lat.name}={lat_.item():.2f} {lat.units}, "
                #     f"{lon.name}={lon_.item():.2f} {lon.units}"
                #     for lat_, lon_ in it
                # ]
                points = [
                    f"{lat_.item():.2f}°, {lon_.item():.2f}°"
                    for lat_, lon_ in it
                ]
            if aspect is None:
                # Integers determine the priority in the case of equal sizes:
                # greater number means higher priority.
                aspect = max(
                    (n_time, 1, "time_series"),
                    (n_vert, 2, "profile"),
                    (n_pts, 3, "points"),
                )[2]
            if aspect not in {"time_series", "profile", "points"}:
                raise ValueError(
                    "'aspect' must be 'time_series', 'profile', 'points', or"
                    " None"
                )
            if aspect == "time_series":
                data = self.to_xarray(encoding=False)[self.name]
                data = data.assign_coords(points=points)

                if vmin is None:
                    vmin = data.min()
                if vmax is None:
                    vmax = data.max()

                kwargs["x"] = time.name
                if n_vert > 1 or (vert is not None and vert.name in data.dims):
                    kwargs.setdefault("row", vert.name)
                if "crs" in data.coords:
                    data = data.drop("crs")
                plot = data.plot.line(vmin=vmin, vmax=vmax, **kwargs)
                if "row" not in kwargs and "col" not in kwargs:
                    for line in plot:
                        line.axes.set_title("Point Time Series")
                if clean_image:
                    plot.axes.set_axis_off()
                    plot.axes.set_title('')
                if save_path:
                    fig = plot[0].figure
                    fig.tight_layout()
                    fig.savefig(save_path, **(save_kwargs or {}))
                return plot
            if aspect == "profile":
                data = self.to_xarray(encoding=False)[self.name]
                data = data.assign_coords(points=points)

                if vmin is None:
                    vmin = data.min()
                if vmax is None:
                    vmax = data.max()

                if vert.attrs.get("positive") == "down":
                    data = data.reindex(
                        indexers={vert.name: data.coords[vert.name][::-1]},
                        copy=False,
                    )
                    data.coords[vert.name] = -data.coords[vert.name]
                    # vert.values = -vert.values[::-1]
                kwargs["y"] = vert.name
                if n_time > 1 or (time is not None and time.name in data.dims):
                    kwargs.setdefault("col", time.name)
                if "crs" in data.coords:
                    data = data.drop("crs")
                plot = data.plot.line(vmin=vmin, vmax=vmax, **kwargs)
                if "row" not in kwargs and "col" not in kwargs:
                    for line in plot:
                        line.axes.set_title("Point Layers")
                if clean_image:
                    plot.axes.set_axis_off()
                    plot.axes.set_title('')
                if save_path:
                    fig = plot[0].figure
                    fig.tight_layout()
                    fig.savefig(save_path, **(save_kwargs or {}))
                return plot

        # Resolving Cartopy features and gridlines:
        if features:
            features = [_CARTOPY_FEATURES[feature] for feature in features]
            if gridlines is None:
                Field._LOG.info("`gridline` turned on")
                gridlines = True
        if gridline_labels is None:
            Field._LOG.info("`gridline_labels` turned off")
            gridline_labels = False
        has_cartopy_items = bool(features or gridlines)

        # Resolving dimensions, coordinates, and coordinate system:
        dims = set()
        if time is not None:
            dims.add(time.name)
        if vert is not None:
            dims.add(vert.name)
        if lat is not None:
            dims.add(lat.name)
            if lat.is_dim:
                kwargs.setdefault("y", lat.name)
        if lon is not None:
            dims.add(lon.name)
            if lon.is_dim:
                kwargs.setdefault("x", lon.name)
        crs = self._domain.crs
        try:
            transform = (
                crs.as_cartopy_projection() if crs is not None else None
            )
        except NotImplementedError:
            # HACK: This is used in the cases where obtaining Cartopy
            # projections is not implemented.
            transform = None
            kwargs.setdefault("x", lon.name)
            kwargs.setdefault("y", lat.name)
        plate = ccrs.PlateCarree

        if len(dims) in {3, 4}:
            if time is not None and time.name in dims and time.size > 1:
                kwargs.setdefault("col", time.name)
            if vert is not None and vert.name in dims and vert.size > 1:
                kwargs.setdefault("row", vert.name)

        # Resolving subplot keyword arguments including `projection`:
        subplot_kwa = {} if subplot_kwargs is None else {**subplot_kwargs}
        if projection is None:
            if has_cartopy_items:
                subplot_kwa["projection"] = projection = plate()
                if transform is None:
                    transform = plate()
            elif isinstance(transform, plate):
                transform = None
            if transform is not None:
                has_cartopy_items = True
                subplot_kwa["projection"] = projection = plate()
        else:
            has_cartopy_items = True
            if isinstance(projection, CoordSystem):
                projection = projection.as_cartopy_projection()
            subplot_kwa["projection"] = projection
            if transform is None:
                transform = plate()
        if subplot_kwa:
            kwargs["subplot_kws"] = subplot_kwa

        # Resolving other keyword arguments including `transform`, `figsize`,
        # and `robust`:
        kwa = {"transform": transform, "figsize": figsize, "robust": robust}
        for name, arg in kwa.items():
            if arg is not None:
                kwargs[name] = arg
        # Creating plot:
        dset = self.to_xarray(encoding=False)
        if "crs" in dset.coords:
            dset = dset.drop("crs")
        # HACK: This should be only:
        # `if self._domain._type is DomainType.GRIDDED:`
        # Checking against `None` is provided temporary for testing.
        if (
            self._domain._type is DomainType.GRIDDED
            or self._domain._type is None
        ):
            data = dset[self.name]

            if vmin is None:
                vmin = data.min().compute()
            if vmax is None:
                vmax = data.max().compute()
            kwargs['vmin'] = vmin
            kwargs['vmax'] = vmax
            #kwargs['levels'] = 10
            print(f'{vmin} / {vmax}')
            if normalize is True:
                data = (vmax - vmin) * ((data - data.min())/(data.max() - data.min())) + vmin
            plot = data.plot(**kwargs)

        elif self._domain._type is DomainType.POINTS:
            data = xr.Dataset(
                data_vars={
                    self.name: dset[self.name],
                    "lat": dset.coords["latitude"],
                    "lon": dset.coords["longitude"],
                }
            )
            kwargs.update(
                {"x": "lon", "y": "lat", "hue": self.name, "zorder": np.inf}
            )

            if vmin is None:
                vmin = data.min().compute()
            if vmax is None:
                vmax = data.max().compute()
            kwargs['vmin'] = vmin
            kwargs['vmax'] = vmax
            plot = data.plot.scatter(**kwargs)
        else:
            raise NotImplementedError(
                "'domain.type' of must be 'DomainType.GRIDDED' or "
                "'DomainType.POINTS'"
            )

        # Adding and modifying axis elements:
        # axes = np.array(getattr(plot, 'axes', plot), copy=False, ndmin=1)

        # Adding gridlines and Cartopy features (borders, coastline, lakes,
        # land, ocean, rivers, or states) to all plot axes:
        if has_cartopy_items:
            axes = np.asarray(plot.axes)
            if features:
                for ax in axes.flat:
                    for feature in features:
                        ax.add_feature(feature)
            if gridlines:
                for ax in axes.flat:
                    ax.gridlines(draw_labels=gridline_labels)

            # NOTE: This is a fix that enables using axis labels and units from
            # the domain, as well as plotting axes labels and ticks when
            # Cartopy transform, projection, or features are used. See:
            # https://stackoverflow.com/questions/35479508/cartopy-set-xlabel-set-ylabel-not-ticklabels
            if (
                (projection is None or isinstance(projection, plate))
                and lat is not None
                and lon is not None
                and not gridline_labels
            ):
                coords = data.coords

                lat_coord = coords[lat.name]
                lat_attrs = lat_coord.attrs
                lat_name = (
                    lat_attrs.get("long_name")
                    or lat_attrs.get("standard_name")
                    or lat_coord.name
                    or "latitude"
                )
                if (lat_units := lat.units) and lat_units != "dimensionless":
                    lat_name = f"{lat_name} [{lat_units}]"
                lat_values = lat.values
                lat_min, lat_max = lat_values.min(), lat_values.max()

                lon_coord = coords[lon.name]
                lon_attrs = lon_coord.attrs
                lon_name = (
                    lon_attrs.get("long_name")
                    or lon_attrs.get("standard_name")
                    or lon_coord.name
                    or "longitude"
                )
                if (lon_units := lon.units) and lon_units != "dimensionless":
                    lon_name = f"{lon_name} [{lon_units}]"
                lon_values = lon.values
                lon_min, lon_max = lon_values.min(), lon_values.max()

                ax = axes.item(0)
                x_ticks = ax.get_xticks()
                x_ticks = x_ticks[(x_ticks >= lon_min) & (x_ticks <= lon_max)]
                y_ticks = ax.get_yticks()
                y_ticks = y_ticks[(y_ticks >= lat_min) & (y_ticks <= lat_max)]

                if axes.ndim == 2:
                    for ax in axes[-1, :].flat:
                        ax.set_xlabel(lon_name)
                        ax.set_xticks(x_ticks)
                    for ax in axes[:, 0].flat:
                        ax.set_ylabel(lat_name)
                        ax.set_yticks(y_ticks)
                else:
                    for ax in axes.flat:
                        ax.set_xlabel(lon_name)
                        ax.set_ylabel(lat_name)
                        ax.set_xticks(x_ticks)
                        ax.set_yticks(y_ticks)

        if clean_image:
            if isinstance(plot.axes, np.ndarray):
                for axis in plot.axes.flat:
                    axis.set_axis_off()
                    axis.set_title('')
            else:
                plot.axes.set_axis_off()
                plot.axes.set_title('')
        if save_path:
            if hasattr(plot, 'fig'):
                fig = plot.fig
            elif hasattr(plot, 'figure'):
                fig = plot.figure
            else:
                raise NotImplementedError()
            fig.tight_layout()
            # fig.tight_layout(pad=0, h_pad=0, w_pad=0)
            fig.savefig(save_path, **(save_kwargs or {}))

        return plot

    def to_image(
        self,
        filepath,
        width,
        height,
        dpi=100,
        format='png',
        transparent=True,
        bgcolor='FFFFFF',
        cmap='RdBu_r',
        projection=None,
        vmin=None,
        vmax=None
    ):
        # NOTE: This method assumes default DPI value.
        f = self
        if self.domain.crs != GeogCS(6371229):
            f = self.to_regular()
#        dpi = plt.rcParams['figure.dpi']
        w, h = width / dpi, height / dpi
        prj = projection
        if prj is not None:
            if prj == '3857':
                # airy1830 = GeogCS(6377563.396, 6356256.909)
                # prj = TransverseMercator(49, -2, 400000, -100000, 0.9996012717, ellipsoid=airy1830)
                prj = WebMercator()

        f.plot(
            figsize=(w, h),
            cmap=cmap,
            add_colorbar=False,
            save_path=filepath,
            save_kwargs={
                'transparent': transparent,
                'pad_inches': 0,
                'dpi': dpi,
                'bbox_inches': 'tight'
            },
            clean_image=True,
            projection=prj,
            vmin=vmin,
            vmax=vmax
        )

    def to_geojson(self, target=None):
        self.load()
        if self.domain.type is DomainType.POINTS:
            if self.latitude.size != 1 or self.longitude.size != 1:
                raise NotImplementedError(
                    "'self.domain' must have exactly 1 point"
                )
            coords = [self.longitude.item(), self.latitude.item()]
            result = {
                "type": "FeatureCollection",
                "units": {self.name: str(self.units)},
                "features": [],
            }
            for time in self.time.values.flat:
                time_ = pd.to_datetime(time).strftime("%Y-%m-%dT%H:%M")
                value = self.sel(time=time_) if self.time.size > 1 else self
                feature = {
                    "type": "Feature",
                    "geometry": {"type": "Point", "coordinates": coords},
                    "properties": {"time": time_, self.name: float(value)},
                }
                result["features"].append(feature)
        elif (
            self.domain.type is DomainType.GRIDDED or self.domain.type is None
        ):
            # HACK: The case `self.domain.type is None` is included to be able
            # to handle undefined domain types temporarily.
            if self.time.size == 1:
                result = {}
            else:
                raise NotImplementedError(
                    f"multiple times are not supported for geojson"
                )
            field = (
                self
                if isinstance(self.domain.crs, GeogCS)
                else self.to_regular()
            )
            axis_names = field.domain._axis_to_name
            lon_min = self.longitude.min().item()
            lat_min = self.latitude.min().item()
            lon_max = self.longitude.max().item()
            lat_max = self.latitude.max().item()              
            grid_x, grid_y = field.domain._infer_resolution()/2.0
            for time in self.time.values.flat:
                time_ = pd.to_datetime(time).strftime("%Y-%m-%dT%H:%M")
                time_data = {
                    "type": "FeatureCollection",
                    "date": time_,
                    "bbox": [
                        lon_min,  # West
                        lat_min,  # South
                        lon_max,  # East
                        lat_max,  # North
                    ],
                    "units": {self.name: str(self.units)},
                    "features": [],
                }
                for lat in field.latitude.values.flat:
                    for lon in field.longitude.values.flat:
                        idx = {
                            axis_names[AxisType.LATITUDE]: lat,
                            axis_names[AxisType.LONGITUDE]: lon,
                        }
                        # if self.time.shape:
                        if self.time.size > 1:
                            idx[axis_names[AxisType.TIME]] = time_
                        # TODO: Check whether this works now:
                        # this gives an error if only 1 time is selected before to_geojson()
                        # 
                        # Polygon:
                        # for each lat/lon we have to define a polygon with lan/lon centered
                        # the cell length depends on the grid resolution (that should be computed)
                        value = field.sel(**idx)
                        lonv = lon.item()
                        latv = lat.item()
                        lon_lower = np.clip(lonv - grid_x, amin=lon_min)
                        lat_upper = np.clip(latv + grid_y, amax=lat_max)
                        lon_upper = np.clip(lonv + grid_x, amax=lon_max)
                        lat_lower = np.clip(latv - grid_y, amin=lat_min)                        
                        feature = {
                            "type": "Feature",
                            "geometry": {
                                "type": "Polygon",
                                "coordinates":[  
                                   [ [lon_lower, lat_upper], [lon_upper, lat_upper],
                                     [lon_upper, lat_lower], [lon_lower, lat_lower],
                                     [lon_lower, lat_upper]
                                   ]
                                ]
                            },
                            "properties": {self.name: float(value)},
                        }
                        time_data["features"].append(feature)
                result = time_data
        else:
            raise NotImplementedError(
                f"'self.domain.type' is {self.domain.type}, which is currently"
                " not supported"
            )

        if target is not None:
            with open(target, mode="w") as file:
                json.dump(result, file, indent=4)

        return result

    @geokube_logging
    def hvplot(self, aspect=None, boxplot=False, **kwargs):
        # NOTE: See https://hvplot.holoviz.org/user_guide/Customization.html
        # for the details on what can be passed with `kwargs`.

        axis_names = self.domain._axis_to_name
        time = self.coords.get(axis_names.get(AxisType.TIME))
        vert = self.coords.get(axis_names.get(AxisType.VERTICAL))
        lat = self.coords.get(axis_names.get(AxisType.LATITUDE))
        lon = self.coords.get(axis_names.get(AxisType.LONGITUDE))

        kwargs.setdefault("widget_location", "bottom")

        dset = self.to_xarray(encoding=False)
        if "crs" in dset.coords:
            dset = dset.drop("crs")
        if (
            vert is not None
            and vert.is_dim
            and vert.attrs.get("positive") == "down"
        ):
            dset = dset.reindex(
                indexers={vert.name: dset.coords[vert.name][::-1]},
                copy=False,
            )
            dset.coords[vert.name] = -dset.coords[vert.name]

        # Considering the case when boxplot is required.
        if boxplot:
            group = kwargs.get("groupby")
            if group == "vertical":
                kwargs["groupby"] = vert.name
            elif group and not isinstance(group, str) and "vertical" in group:
                group = list(group)
                idx = group.index("vertical")
                group[idx] = vert.name
                kwargs["groupby"] = group

            data = dset[self.name]

            if self._domain._type is DomainType.POINTS:
                with np.nditer((lat.values, lon.values)) as it:
                    points = [
                        f"{lat_.item():.2f}°, {lon_.item():.2f}°"
                        for lat_, lon_ in it
                    ]
                data = data.assign_coords(points=points)

            return data.hvplot.box(y=self.name, **kwargs)

        # Working with `DomainType.POINTS`.
        if self._domain._type is DomainType.POINTS:
            n_pts = lat.size
            n_time = time.size if (time is not None and time.is_dim) else 0
            n_vert = vert.size if (vert is not None and vert.is_dim) else 0
            if vert is None or vert.is_dim:
                # NOTE: Case when vertical is given as a profile.
                vals = (lat.values, lon.values)
                with np.nditer(vals) as it:
                    # NOTE: This approach might result in an error when the
                    # string representations of multiple points are equal.
                    # points = [
                    #     f'{lat.name}={lat_.item():.2f} {lat.units}, '
                    #     f'{lon.name}={lon_.item():.2f} {lon.units}'
                    #     for lat_, lon_ in it
                    # ]
                    points = [
                        f"{lat_.item():.2f}°, {lon_.item():.2f}°"
                        for lat_, lon_ in it
                    ]
                if aspect is None:
                    # Integers determine the priority in the case of equal:
                    # sizes greater number means higher priority.
                    aspect = max(
                        (n_time, 1, "time_series"),
                        (n_vert, 2, "profile"),
                        (n_pts, 3, "points"),
                    )[2]
            else:
                # NOTE: Case when vertical is given as points.
                # NOTE: This approach might result in an error when the
                # string representations of multiple points are equal.
                dset.coords[vert.name] = -dset.coords[vert.name]
                vals = (lat.values, lon.values, dset.coords[vert.name].values)
                with np.nditer(vals) as it:
                    points = [
                        f"{lat_.item():.2f}°, {lon_.item():.2f}° "
                        f"{vert_.item():.2f} {vert.units}"
                        for lat_, lon_, vert_ in it
                    ]
                if aspect is None:
                    # Integers determine the priority in the case of equal:
                    # sizes greater number means higher priority.
                    aspect = max(
                        (n_time, 1, "time_series"),
                        (n_pts, 2, "points"),
                    )[2]
            if aspect == "time_series":
                data = dset[self.name]
                data = data.assign_coords(points=points)
                kwargs.update({"geo": False, "tiles": False})
                return data.hvplot(x=time.name, by="points", **kwargs)
            if aspect == "profile":
                data = dset[self.name]
                data = data.assign_coords(points=points)
                kwargs.update({"geo": False, "tiles": False})
                return data.hvplot(y=vert.name, by="points", **kwargs)
            if aspect == "points":
                data_vars = {
                    self.name: dset[self.name],
                    "lat": dset.coords["latitude"],
                    "lon": dset.coords["longitude"],
                }
                if (vert is not None) and (not vert.is_dim):
                    data_vars["vert"] = dset.coords[vert.name]
                    if "groupby" not in kwargs:
                        kwargs["groupby"] = sorted(
                            self.coords.keys() - {"latitude", "longitude"}
                        )
                data = xr.Dataset(data_vars=data_vars)
                return data.hvplot.scatter(
                    x="lon",
                    y="lat",
                    c=self.name,
                    cmap=kwargs.pop("cmap", "coolwarm"),
                    colorbar=kwargs.pop("colorbar", True),
                    **kwargs,
                )
            raise ValueError(
                "'aspect' must be 'time_series', 'profile', 'points', or None"
            )

        # Working with `DomainType.GRIDDED`.
        # HACK: This should be only:
        # `if self._domain._type is DomainType.GRIDDED:`
        # Checking against `None` is provided temporary for testing.
        if (
            self._domain._type is DomainType.GRIDDED
            or self._domain._type is None
        ):
            crs = self._domain.crs
            if crs is not None:
                try:
                    crs = crs.as_cartopy_projection()
                except NotImplementedError:
                    # HACK: This is used in the cases where obtaining Cartopy
                    # projections is not implemented.
                    crs = None
                    kwargs.setdefault("x", lon.name)
                    kwargs.setdefault("y", lat.name)

            proj = kwargs.get("projection")
            if isinstance(proj, CoordSystem):
                kwargs["projection"] = proj = proj.as_cartopy_projection()

            if aspect is not None:
                if aspect == "time_series":
                    kwargs.update({"x": time.name, "y": self.name})
                    kwargs.update({"geo": False, "tiles": False})
                elif aspect == "profile":
                    kwargs.update({"x": self.name, "y": vert.name})
                    kwargs.update({"geo": False, "tiles": False})
                if crs is not None and not isinstance(crs, ccrs.PlateCarree):
                    dset = self.to_regular().to_xarray(encoding=False)

            plot_call = dset[self.name].hvplot

            if lat is not None and lat.is_dim:
                kwargs.setdefault("y", lat.name)
            if lon is not None and lon.is_dim:
                kwargs.setdefault("x", lon.name)

            if (
                crs is None
                and lat is not None
                and lon is not None
                and (lat.size > 1 or lon.size > 1)
                and aspect is None
            ):
                plot_call = plot_call.quadmesh
                kwargs.setdefault("rasterize", True)
                kwargs.setdefault("project", True)

            if (
                not (
                    (proj is None or isinstance(proj, ccrs.PlateCarree))
                    and (crs is None or isinstance(crs, ccrs.PlateCarree))
                )
                and "x" not in kwargs
                and "y" not in kwargs
                and lat is not None
                and lon is not None
            ):
                plot_call = plot_call.quadmesh
                kwargs["crs"] = crs
                kwargs.setdefault("rasterize", True)
                kwargs.setdefault("project", True)
                lat_name = lat.attrs.get("long_name", lat.name)
                if (lat_units := lat.attrs.get("units")) is not None:
                    lat_name = f"{lat_name} ({lat_units})"
                lon_name = lon.attrs.get("long_name", lon.name)
                if (lon_units := lon.attrs.get("units")) is not None:
                    lon_name = f"{lon_name} ({lon_units})"
                kwargs.update({"xlabel": lon_name, "ylabel": lat_name})

            # TODO: Consider improving the logic (handling conditions).
            if (
                kwargs.get("tiles")
                and lat is not None
                and lon is not None
                and kwargs.get("x") == lon.name
                and kwargs.get("y") == lat.name
                and aspect is None
            ):
                lat_name = lat.attrs.get("long_name", lat.name)
                if (lat_units := lat.attrs.get("units")) is not None:
                    lat_name = f"{lat_name} ({lat_units})"
                lon_name = lon.attrs.get("long_name", lon.name)
                if (lon_units := lon.attrs.get("units")) is not None:
                    lon_name = f"{lon_name} ({lon_units})"
                kwargs.update({"xlabel": lon_name, "ylabel": lat_name})

            return plot_call(**kwargs)

        raise NotImplementedError(
            "'domain.type' must be 'DomainType.GRIDDED' or 'DomainType.POINTS'"
        )

    @geokube_logging
    def box_plot(self, by=None, orientation="vertical", **kwargs):
        # NOTE: `kwargs` are passed directly or in a slightly modified form
        # to `plotly.express.box`. For more details, see the official
        # documentation:
        # * https://plotly.github.io/plotly.py-docs/generated/plotly.express.box.html
        # * https://plotly.com/python/box-plots/

        axis_names = self.domain._axis_to_name
        time = self.coords.get(axis_names.get(AxisType.TIME))
        vert = self.coords.get(axis_names.get(AxisType.VERTICAL))
        lat = self.coords.get(axis_names.get(AxisType.LATITUDE))
        lon = self.coords.get(axis_names.get(AxisType.LONGITUDE))

        if by is None:
            by_ = None
        elif by == "points" and self._domain._type is DomainType.POINTS:
            by_ = "points"
        else:
            if by in (dim_types := {ax.value[0] for ax in axis_names.keys()}):
                by_ = getattr(self, by).name
            elif by not in (dim_names := {*axis_names.values()}):
                raise ValueError(
                    "'by' must be 'None' or one of the following: "
                    f"{sorted(dim_types | dim_names)}"
                )

        if orientation in {"h", "horizontal"}:
            kwargs.setdefault("x", self.name)
            if by_:
                kwargs.setdefault("y", by_)
        elif orientation in {"v", "vertical"}:
            kwargs.setdefault("y", self.name)
            if by_:
                kwargs.setdefault("x", by_)
        else:
            raise ValueError(
                "'orientation' must be either 'horizontal' ('h') or "
                "'vertical' ('v')"
            )

        dset = self.to_xarray(encoding=False)
        if "crs" in dset.coords:
            dset = dset.drop("crs")
        if (
            vert is not None
            and vert.is_dim
            and vert.attrs.get("positive") == "down"
        ):
            dset = dset.reindex(
                indexers={vert.name: dset.coords[vert.name][::-1]},
                copy=False,
            )
            dset.coords[vert.name] = -dset.coords[vert.name]

        darr = dset[self.name]
        if self._domain._type is DomainType.POINTS:
            with np.nditer((lat.values, lon.values)) as it:
                points = [
                    f"{lat_.item():.2f}°, {lon_.item():.2f}°"
                    for lat_, lon_ in it
                ]
                darr = darr.assign_coords(points=points)

        df = darr.to_dataframe()
        df_ = df.index.to_frame()
        df_.index = np.arange(df_.shape[0])
        for col_name in df.columns.to_numpy().flat:
            df_[col_name] = df[col_name].to_numpy()
        af = kwargs.get("animation_frame")
        if af and af == time.name:
            kwargs["animation_frame"] = animation_name = af + "_"
            df_[animation_name] = df_[af].astype(str)

        fig = px.box(df_, **kwargs)

        if af:
            min_, max_ = np.nanmin(self.values), np.nanmax(self.values)
            margin = 0.05 * (np.nanmax(self.values) - np.nanmin(self.values))
            bounds = [min_ - margin, max_ + margin]
            name = f"{'x' if orientation[0] == 'h' else 'y'}axis_range"
            kwa = {name: bounds}
            fig.update_layout(**kwa)
            # for f in fig.frames:
            #     f.layout.update(**kwa)

        return fig

    @geokube_logging
    def to_xarray(self, encoding=True) -> xr.Dataset:
        data_vars = {}
        var_name = self.ncvar if encoding else self.name

        data_vars[var_name] = super().to_xarray(
            encoding
        )  # use Variable to_array

        coords = self.domain.aux_coords
        if coords:
            if encoding:
                coords_names = " ".join(
                    [self.domain.coords[x].ncvar for x in coords]
                )
            else:
                coords_names = " ".join(
                    [self.domain.coords[x].name for x in coords]
                )
            data_vars[var_name].encoding["coordinates"] = coords_names

        coords = self.domain.to_xarray(encoding)

        crs_name = self.domain.grid_mapping_name
        data_vars[var_name].encoding["grid_mapping"] = crs_name

        if self.cell_methods is not None:
            data_vars[var_name].attrs["cell_methods"] = str(self.cell_methods)

        if self._ancillary is not None:
            for a in self.ancillary:
                data_vars[a] = a.to_xarray(encoding)

        # NOTE: a workaround for keeping domaintype
        # Issue: https://github.com/geokube/geokube/issues/147
        # If saved in .nc file, domain_type should be converted to str
        if (domain_type := self.domain.type) is not None and (not encoding):
            data_vars[var_name].attrs["__geo_domtype"] = domain_type

        return xr.Dataset(data_vars=data_vars, coords=coords)

    def persist(self, path=None) -> str:
        if path is None:
            path = os.path.join(
                tempfile.gettempdir(), f"{str(uuid.uuid4())}.nc"
            )
        if os.path.isdir(path):
            path = os.path.join(path, f"{str(uuid.uuid4())}.nc")
        if not path.endswith(".nc"):
            self._LOG.warn(
                f"Provided persistance path: `{path}` has not `.nc` extension."
                " Adding automatically!"
            )
            warnings.warn(
                f"Provided persistance path: `{path}` has not `.nc` extension."
                " Adding automatically!"
            )
            path = path + ".nc"
        self.to_netcdf(path)
        return path

    def to_dict(self):
        description = None
        if self._mapping is not None and self.ncvar in self._mapping:
            var_map = self._mapping[self.ncvar]
            if "description" in var_map:
                description = var_map["description"]
            elif "name" in var_map:
                description = var_map["name"]
        if description is None:
            description = self.properties.get(
                "description", self.properties.get("long_name")
            )
        return {
            "units": str(self.units),
            "description": description,
        }

    @classmethod
    @geokube_logging
    def from_xarray(
        cls,
        ds: xr.Dataset,
        ncvar: str,
        id_pattern: Optional[str] = None,
        mapping: Optional[Mapping[str, Mapping[str, str]]] = None,
        copy=False,
    ):
        if not isinstance(ds, xr.Dataset):
            raise TypeError(
                f"Expected type `xarray.Dataset` but provided `{type(ds)}`"
            )
        ds = convert_cftimes_to_numpy(ds)
        domain = Domain.from_xarray(
            ds, ncvar=ncvar, id_pattern=id_pattern, copy=copy, mapping=mapping
        )
        da = ds[ncvar].copy(copy)  # TODO: TO CHECK
        cell_methods = CellMethod.parse(da.attrs.pop("cell_methods", None))
        var = Variable.from_xarray(da, id_pattern, mapping=mapping)

        name = Variable._get_name(da, mapping=mapping, id_pattern=id_pattern)

        # We need to update `encoding` of var, as `Variable` doesn't contain `name`
        var.encoding.update(name=da.encoding.get("name", ncvar))
        # TODO ancillary variables
        field = Field(
            name=name,
            data=var.data,
            dims=var.dims,
            units=var.units,
            properties=var.properties,
            encoding=var.encoding,
            cell_methods=cell_methods,
            coords=domain,
        )
        field._id_pattern = id_pattern
        field._mapping = mapping
        field._domain._calculate_missing_lat_and_lon()
        return field

    @staticmethod
    def _update_coordinates(da: xr.DataArray, coords):
        if coords is None or len(coords) == 0:
            return
        if "coordinates" in da.attrs:
            da.attrs["coordinates"] = " ".join(
                chain([da.attrs["coordinates"]], coords)
            )
        elif "coordinates" in da.encoding:
            da.encoding["coordinates"] = " ".join(
                chain([da.encoding["coordinates"]], coords)
            )
        else:
            da.encoding["coordinates"] = " ".join(coords)



import logging
import os
import timeit
from numbers import Number

import numpy as np
import shutil

RES_PATH = os.path.join("tests", "resources", "__res.nc")
RES_DIR = os.path.join("tests", "resources", "__res")


def compare_dicts(d1, d2, exclude_d1=None, exclude_d2=None):
    exclude_d1 = (
        list(np.array(exclude_d1, ndmin=1)) if exclude_d1 is not None else []
    )
    exclude_d2 = (
        list(np.array(exclude_d2, ndmin=1)) if exclude_d2 is not None else []
    )
    for ek in set(d1.keys()):
        if ek in exclude_d1:
            continue
        assert ek in d2, f"{ek} not present in d2"
        if d1[ek] is None:
            assert d2[ek] is None
        if isinstance(d1[ek], Number) and np.isnan(d1[ek]):
            assert np.isnan(d2[ek])
        else:
            assert ek in d2
            assert d1[ek] == d2[ek], f"Key: {ek}. {d1[ek]} != {d2[ek]}"
    for ek in set(d2.keys()) - set(d1.keys()) - set(exclude_d2):
        assert ek in d1, f"{ek} not present in d1"
        if d2[ek] is None:
            assert d1[ek] is None
        if isinstance(d2[ek], Number) and np.isnan(d2[ek]):
            assert np.isnan(d1[ek])
        else:
            assert d1[ek] == d2[ek], f"Key: {ek}. {d1[ek]} != {d2[ek]}"


def clear_test_res():
    for f in [RES_DIR, RES_DIR]:
        try:
            shutil.rmtree(f)
        except:
            pass


class TimeCounter:
    def __init__(self, print=False, log=False):
        self.__print = print
        self.__log = log
        if self.__log:
            self.logger = logging.getLogger("TimeCounter")

    def __enter__(self):
        self.__start_time = timeit.timeit()
        return self

    def __exit__(self, type, value, traceback):
        self.execution_time = timeit.timeit() - self.__start_time
        if self.__print:
            print(f"Execution took: {self.execution_time} msec")
        if self.__log:
            self.logger.info(f"Execution took: {self.execution_time} msec")

    @property
    def exec_time(self):
        return self.execution_time



import geopandas as gpd
import pytest


@pytest.fixture
def it_shape_100_km():
    path = "tests//resources//italy_shapefile//it_100km.shp"
    yield gpd.read_file(path).to_crs(epsg=4326)["geometry"]



import numpy as np
import pandas as pd
import pytest

import geokube.core.coord_system as crs
from geokube.core.axis import Axis, AxisType
from geokube.core.bounds import Bounds1D, BoundsND
from geokube.core.coordinate import Coordinate, CoordinateType
from geokube.core.domain import Domain, DomainType
from geokube.core.unit import Unit
from geokube.core.variable import Variable
from tests import compare_dicts
from tests.fixtures import *


def test_from_xarray_rotated_pole_wso(era5_rotated_netcdf):
    domain = Domain.from_xarray(era5_rotated_netcdf, ncvar="W_SO")
    # TODO: domaintype is currently not set
    # assert domain.type is DomainType.GRIDDED
    assert domain.crs == crs.RotatedGeogCS(
        grid_north_pole_latitude=47, grid_north_pole_longitude=-168
    )
    assert "time" in domain
    assert "latitude" in domain
    assert "longitude" in domain
    assert "depth" in domain
    assert "grid_latitude" in domain
    assert "grid_longitude" in domain
    assert domain.vertical.has_bounds
    assert isinstance(domain.vertical.bounds["soil1_bnds"], Bounds1D)

    res = domain.to_xarray(encoding=False)
    assert isinstance(res, xr.core.coordinates.DatasetCoordinates)
    assert "time" in res
    assert "latitude" in res
    assert "longitude" in res
    assert "depth" in res
    assert "grid_latitude" in res
    assert "grid_longitude" in res


def test_from_xarray_rotated_pole_tmin2m(era5_rotated_netcdf):
    domain = Domain.from_xarray(era5_rotated_netcdf, ncvar="TMIN_2M")
    # TODO: domaintype is currently not set
    # assert domain.type is DomainType.GRIDDED
    assert domain.crs == crs.RotatedGeogCS(
        grid_north_pole_latitude=47, grid_north_pole_longitude=-168
    )
    assert "time" in domain
    assert "latitude" in domain
    assert "longitude" in domain
    assert "height" in domain
    assert "grid_latitude" in domain
    assert "grid_longitude" in domain

    res = domain.to_xarray(encoding=True)
    assert isinstance(res, xr.core.coordinates.DatasetCoordinates)
    assert "time" in res
    assert "lat" in res
    assert "lon" in res
    assert "height_2m" in res
    assert "rlat" in res
    assert "rlon" in res


def test_from_xarray_curvilinear_grid(nemo_ocean_16):
    domain = Domain.from_xarray(nemo_ocean_16, ncvar="vt")
    assert "time" in domain
    assert "latitude" in domain
    assert "longitude" in domain
    assert "depthv" in domain  # domain.depthv.attrs['name'] is `depthv`
    assert domain["depthv"].units == Unit("m")
    assert "x" not in domain
    assert "y" not in domain
    assert domain.time.has_bounds
    assert isinstance(domain.time.bounds["time_counter_bounds"], Bounds1D)
    assert domain.vertical.has_bounds
    assert isinstance(domain.vertical.bounds["depthv_bounds"], Bounds1D)
    assert domain.latitude.has_bounds
    assert isinstance(domain.latitude.bounds["bounds_lat"], BoundsND)
    assert domain.longitude.has_bounds
    assert isinstance(domain.longitude.bounds["bounds_lon"], BoundsND)


def test_from_xarray_regular_latlon(era5_globe_netcdf):
    domain = Domain.from_xarray(era5_globe_netcdf, ncvar="tp")
    res = domain.to_xarray()
    assert "latitude" in domain
    assert "latitude" in res
    assert "units" in res["latitude"].attrs
    compare_dicts(
        res["latitude"].attrs,
        era5_globe_netcdf["latitude"].attrs,
        exclude_d1="standard_name",
    )
    compare_dicts(
        res["latitude"].encoding,
        era5_globe_netcdf["latitude"].encoding,
        exclude_d1=["name", "_FillValue"],
        exclude_d2="_FillValue",
    )
    assert res["latitude"].encoding["_FillValue"] is None
    assert res["latitude"].encoding["name"] == "latitude"
    assert "longitude" in domain
    assert "longitude" in res
    assert "units" in res["longitude"].attrs
    compare_dicts(
        res["longitude"].encoding,
        era5_globe_netcdf["longitude"].encoding,
        exclude_d1=["name", "_FillValue"],
        exclude_d2="_FillValue",
    )
    assert res["latitude"].encoding["_FillValue"] is None
    compare_dicts(
        res["longitude"].attrs,
        era5_globe_netcdf["longitude"].attrs,
        exclude_d1="standard_name",
    )
    assert res["longitude"].encoding["name"] == "longitude"
    assert "time" in domain
    assert "time" in res
    assert "units" in res["time"].encoding
    assert "calendar" in res["time"].encoding
    compare_dicts(
        res["time"].attrs,
        era5_globe_netcdf["time"].attrs,
        exclude_d1="standard_name",
    )
    compare_dicts(
        res["time"].encoding,
        era5_globe_netcdf["time"].encoding,
        exclude_d1=["name", "_FillValue"],
    )
    assert res["time"].encoding["_FillValue"] is None
    assert res["time"].encoding["name"] == "time"
    assert domain.crs == crs.RegularLatLon()
    assert "crs_latitude_longitude" in res
    assert res["crs_latitude_longitude"].attrs == {
        "semi_major_axis": 6371229.0,
        "grid_mapping_name": "latitude_longitude",
    }


def test_regular_lat_lon_domain_from_coords_dict():
    coords = {
        "lat": np.arange(1, 10),
        "lon": np.arange(6, 50),
        "height_2m": np.arange(4, 14),
        "time": pd.date_range("01-01-2001", "10-01-2001", freq="1D"),
    }
    dims = ("time", "lat", "lon", "height_2m")
    domain = Domain._make_domain_from_coords_dict_dims_and_crs(coords, dims)
    assert domain.crs == crs.RegularLatLon()
    assert Axis("time") in domain
    assert AxisType.TIME in domain
    assert domain[AxisType.TIME].is_dim
    assert Axis("lat") in domain
    assert AxisType.LATITUDE in domain
    assert domain[Axis("latitude")].is_dim
    assert Axis("lon") in domain
    assert AxisType.LONGITUDE in domain
    assert domain["lon"].is_dim
    assert Axis("vertical") in domain
    assert AxisType.VERTICAL in domain
    assert domain[Axis("height_2m")].is_dim

    coords = {
        AxisType.LONGITUDE: np.arange(1, 10),
        AxisType.LATITUDE: np.arange(6, 50),
        "height_2m": np.arange(4, 14),
        "time": pd.date_range("01-01-2001", "10-01-2001", freq="1D"),
    }
    dims = (AxisType.TIME, Axis("lat"), "lon", "height_2m")
    domain = Domain._make_domain_from_coords_dict_dims_and_crs(coords, dims)
    assert domain.crs == crs.RegularLatLon()
    assert Axis("time") in domain
    assert AxisType.TIME in domain
    assert Axis("lat") in domain
    assert AxisType.LATITUDE in domain
    assert Axis("lon") in domain
    assert AxisType.LONGITUDE in domain
    assert Axis("vertical") in domain
    assert AxisType.VERTICAL in domain


def test_curvilinear_lat_lon_domain_from_coords_dict():
    dims = ("time", "x", "y")
    coords = {
        "lat": ((Axis("x"), Axis("y")), np.array([[0, 1], [1, 2]])),
        "lon": ((Axis("x"), Axis("y")), np.array([[0, 1], [1, 2]])),
        "time": pd.date_range("01-01-2001", "10-01-2001", freq="1D"),
    }
    domain = Domain._make_domain_from_coords_dict_dims_and_crs(
        coords, dims, crs=crs.CurvilinearGrid()
    )
    assert domain.crs == crs.CurvilinearGrid()
    assert Axis("lat") in domain
    assert AxisType.LATITUDE in domain
    assert not domain[AxisType.LATITUDE].is_dim
    assert Axis("lon") in domain
    assert AxisType.LONGITUDE in domain
    assert not domain[AxisType.LONGITUDE].is_dim
    assert Axis("x") not in domain
    assert AxisType.X not in domain
    assert Axis("y") not in domain
    assert AxisType.Y not in domain

    dims = (Axis("time"), AxisType.X, AxisType.Y)
    coords = {
        "lat": ((AxisType.X, AxisType.Y), np.array([[0, 1], [1, 2]])),
        "lon": (("x", "y"), np.array([[0, 1], [1, 2]])),
        "time": pd.date_range("01-01-2001", "10-01-2001", freq="1D"),
    }
    domain = Domain._make_domain_from_coords_dict_dims_and_crs(
        coords, dims, crs=crs.CurvilinearGrid()
    )
    assert domain.crs == crs.CurvilinearGrid()
    assert Axis("lat") in domain
    assert AxisType.LATITUDE in domain
    assert not domain["lat"].is_dim
    assert Axis("lon") in domain
    assert AxisType.LONGITUDE in domain
    assert not domain["lon"].is_dim
    assert Axis("x") not in domain
    assert AxisType.X not in domain
    assert Axis("y") not in domain
    assert AxisType.Y not in domain


def test_rotated_pole_lat_lon_domain_from_coords_dict():
    dims = ("time", AxisType.X, AxisType.Y)
    coords = {
        "lat": ((Axis("rlat"), Axis("rlon")), np.array([[0, 1], [1, 2]])),
        "lon": ((Axis("rlat"), Axis("rlon")), np.array([[0, 1], [1, 2]])),
        "rlat": np.array([6, 7]),
        "rlon": np.array([9, 8]),
        "time": pd.date_range("01-01-2001", "10-01-2001", freq="1D"),
    }
    domain = Domain._make_domain_from_coords_dict_dims_and_crs(
        coords, dims, crs=crs.RotatedGeogCS(12345, 5678)
    )
    assert isinstance(domain.crs, crs.RotatedGeogCS)
    assert Axis("lat") in domain
    assert AxisType.LATITUDE in domain
    assert not domain[AxisType.LATITUDE].is_dim
    assert Axis("lon") in domain
    assert AxisType.LONGITUDE in domain
    assert not domain[AxisType.LONGITUDE].is_dim
    assert Axis("rlat") in domain
    assert AxisType.X in domain
    assert domain[AxisType.X].is_dim
    assert Axis("y") in domain
    assert AxisType.Y in domain
    assert domain[AxisType.Y].is_dim


def test_regular_lat_lon_domain_with_bounds_and_properties_from_dict():
    time_bounds = np.stack(
        (
            pd.date_range("01-01-2001", "10-01-2001", freq="1D"),
            pd.date_range("02-01-2001", "11-01-2001", freq="1D"),
        )
    ).T
    coords = {
        "lat": np.arange(1, 10),
        "lon": np.arange(6, 50),
        "height_2m": np.arange(4, 14),
        "time": (
            ("time",),
            pd.date_range("01-01-2001", "10-01-2001", freq="1D"),
            time_bounds,
            None,
            {"some_prop": "prop_val"},
        ),
    }
    dims = ("time", "lat", "lon", "height_2m")
    domain = Domain._make_domain_from_coords_dict_dims_and_crs(coords, dims)
    assert "time" in domain
    assert domain[AxisType.TIME].has_bounds
    assert "some_prop" in domain[Axis("Time")].properties
    assert domain[Axis("Time")].properties["some_prop"] == "prop_val"


def test_skip_redundand_coord_from_coordinates_string_in_encoding(
    era5_globe_netcdf,
):
    era5_globe_netcdf["tp"].encoding["coordinates"] = "time latitude"
    with pytest.warns(
        UserWarning,
        match=r"Coordinate (time|latitude) was already defined as dimension!",
    ):
        Domain.from_xarray(era5_globe_netcdf, ncvar="tp")


def test_skip_redundand_coord_from_coordinates_string_in_attrs(
    era5_globe_netcdf,
):
    era5_globe_netcdf["tp"].attrs["coordinates"] = "time latitude"
    with pytest.warns(
        UserWarning,
        match=r"Coordinate (time|latitude) was already defined as dimension!",
    ):
        Domain.from_xarray(era5_globe_netcdf, ncvar="tp")


def test_skip_coordinate_from_coordinates_string_if_not_present_in_dataset(
    era5_globe_netcdf,
):
    era5_globe_netcdf["tp"].attrs["coordinates"] = "time not_exsiting"
    with pytest.warns(
        UserWarning,
        match=r"Coordinate not_exsiting does not exist in the dataset!",
    ):
        Domain.from_xarray(era5_globe_netcdf, ncvar="tp")


def test_using_geo_domtype_attribute_for_domain_type(era5_globe_netcdf):
    era5_globe_netcdf["tp"].attrs["__geo_domtype"] = "gridded"
    domain = Domain.from_xarray(era5_globe_netcdf, ncvar="tp")
    assert domain.type is DomainType.GRIDDED


def test_to_dict_store_proper_keys(era5_globe_netcdf):
    details = Domain.from_xarray(era5_globe_netcdf, ncvar="tp").to_dict()
    assert isinstance(details, dict)
    assert "crs" in details
    assert "coordinates" in details


def test_to_dict_store_reg_crs_with_names_and_attributes(era5_globe_netcdf):
    details = Domain.from_xarray(era5_globe_netcdf, ncvar="tp").to_dict()
    crs = details["crs"]
    assert isinstance(crs, dict)
    assert crs["name"] == "latitude_longitude"
    assert "semi_major_axis" in crs
    assert crs["semi_major_axis"] == 6371229.0
    assert "semi_minor_axis" in crs
    assert crs["semi_minor_axis"] == 6371229.0
    assert "inverse_flattening" in crs
    assert crs["inverse_flattening"] == 0.0
    assert "longitude_of_prime_meridian" in crs
    assert crs["longitude_of_prime_meridian"] == 0.0


def test_to_dict_store_rotated_crs_with_names_and_attributes(
    era5_rotated_netcdf,
):
    details = Domain.from_xarray(era5_rotated_netcdf, ncvar="W_SO").to_dict()
    crs = details["crs"]
    assert isinstance(crs, dict)
    assert crs["name"] == "rotated_latitude_longitude"
    assert "grid_north_pole_latitude" in crs
    assert crs["grid_north_pole_latitude"] == 47.0
    assert "grid_north_pole_longitude" in crs
    assert crs["grid_north_pole_longitude"] == -168.0
    assert "north_pole_grid_longitude" in crs
    assert crs["north_pole_grid_longitude"] == 0
    assert "ellipsoid" in crs
    assert crs["ellipsoid"] is None


@pytest.mark.skip(
    "Invalidate as in the current version, Domain does not contain data values"
)
def test_to_dict_store_coords_reg_crs(era5_globe_netcdf):
    details = Domain.from_xarray(era5_globe_netcdf, ncvar="tp").to_dict()
    coords = details["coordinates"]
    assert len(coords) == 3
    assert isinstance(coords, dict)
    assert set(coords.keys()) == {"time", "latitude", "longitude"}
    for k in coords.values():
        assert "values" in k
        assert isinstance(k["values"], list)

@pytest.mark.skip(
    "Invalidate as in the current version, Coordinate does not contain data values"
)
def test_to_dict_store_coords_rot_crs(era5_rotated_netcdf):
    details = Domain.from_xarray(era5_rotated_netcdf, ncvar="W_SO").to_dict()
    coords = details["coordinates"]
    assert len(coords) == 6
    assert isinstance(coords, dict)
    assert set(coords.keys()) == {
        "depth",
        "time",
        "grid_latitude",
        "grid_longitude",
        "latitude",
        "longitude",
    }
    for k in coords.values():
        assert "values" in k
        assert isinstance(k["values"], list)



import datetime
import os
import re
from typing import Any, Mapping, NoReturn, Optional, Tuple

import numpy as np
import pandas as pd


def _rev_format_to_glob_pattern(val: str):
    return os.sep.join(
        map(lambda s: re.sub("{.*}", "*", s), val.split(os.sep))
    )


def convert_cftimes_to_numpy(obj):
    for key in obj.coords:
        if obj[key].dtype == np.dtype("O"):
            try:
                obj[key] = obj.indexes[key].to_datetimeindex()
            except AttributeError:
                pass
    return obj



from datetime import datetime
import re
from collections.abc import Iterable
from itertools import product
from numbers import Number
from string import Formatter, Template
from typing import Any, Hashable, Mapping, Union

import numpy as np
import xarray as xr
from dask import is_dask_collection

from ..core.axis import Axis


def convert_cftimes_to_numpy(obj):
    for key in obj.coords:
        if obj[key].dtype == np.dtype("O"):
            try:
                obj[key] = obj.indexes[key].to_datetimeindex()
            except AttributeError:
                pass
            except KeyError:
                # NOTE: we skipp non-indexable cooridnates, like time_bnds
                pass            
    return obj


def are_dims_compliant(provided_shape, expected_shape):
    return provided_shape == expected_shape


def trim_key(mapping: dict, exclude: list):
    return {
        k: v for k, v in mapping.items() if k not in np.array(exclude, ndmin=1)
    }


def cast_attr_to_valid(
    attrs: Mapping[Hashable, Any]
) -> Mapping[Hashable, Union[str, Number, np.ndarray, list, tuple]]:
    return {
        k: str(v) if isinstance(v, np.dtype) else v for k, v in attrs.items()
    }


def clear_attributes(d: Union[xr.Dataset, xr.DataArray], attrs):
    attrs = np.array(attrs, ndmin=1)
    if isinstance(d, xr.DataArray):
        for a in attrs:
            d.attrs.pop(a, None)
    elif isinstance(d, xr.Dataset):
        for dv in d.data_vars:
            for a in attrs:
                d[dv].attrs.pop(a, None)


def get_indexer_list_product(indexers):
    dim_names = list(indexers.keys())
    cart_prod = product(*[np.array(v, ndmin=1) for v in indexers.values()])
    return [{d: c for d, c in zip(dim_names, c)} for c in cart_prod]


def get_not_nones(*args) -> list:
    return list(filter(lambda x: x is not None, args))


def is_atleast_one_not_none(*args, **kwargs):
    for v in args:
        if v is not None:
            return True
    for v in kwargs.values():
        if v is not None:
            return True
    return False


def assert_exactly_one_arg(**kwargs):
    provided = False
    for v in kwargs.values():
        if provided and v is not None:
            raise ValueError(
                "Those arguments cannot be passed at ones:"
                f" {list(kwargs.keys())}"
            )
        if not provided and v is not None:
            provided = True
    if not provided:
        raise ValueError(
            f"Exactly one argument must be provided: {list(kwargs.keys())}"
        )


def _agnostic_operator(method, value):
    if is_dask_collection(value):
        return method(value.compute())
    return method(value)


def agnostic_max(value):
    return _agnostic_operator(np.max, value)


def agnostic_min(value):
    return _agnostic_operator(np.min, value)


def agnostic_diff(value):
    return _agnostic_operator(np.diff, value)


def agnostic_unique(value):
    return _agnostic_operator(np.unique, value)


def is_time_combo_key(indexer: dict):
    return (
        isinstance(indexer, dict)
        and (indexer is not None)
        and (
            ("hour" in indexer.keys())
            or ("day" in indexer.keys())
            or ("month" in indexer.keys())
            or ("year" in indexer.keys())
        )
    )


def make_compatible_data(data):
    if isinstance(data, np.ndarray):
        # Necessery as np.ndarray is sometimes treated as memview class, which cannot be processed by xarray!
        return np.asarray(data)
    return data


def slice_common_part(s1, s2):
    start = np.max([np.min([s1.start, s1.stop]), np.min([s2.start, s2.stop])])
    stop = np.min([np.max([s1.start, s1.stop]), np.max([s2.start, s2.stop])])
    return slice(start, stop)


def list_to_slice_or_array(val_list):
    if isinstance(val_list, slice):
        return val_list
    elif isinstance(val_list, set):
        return list(val_list)
    elif isinstance(val_list, Iterable):
        if len(val_list) == 2:
            return val_list
        arr = np.array(list(val_list))
        if len(np.unique(df := np.diff(arr))) == 1:
            return slice(arr[0], arr[-1] + df[0], df[0])
        return arr
    return val_list


def parse_cell_measures_string(cell_measures_str) -> dict:
    return {
        vi.strip(): vi.strip()
        for v in re.findall(r"[a-z]*[:].?[a-z]*", cell_measures_str)
        for vi in v.split(":")
    }


def _is_time_part_name(name: str) -> bool:
    return name == "year" or name == "month" or name == "day" or name == "hour"


def find_slice_dict_and_change_to_slice(dict_val: Mapping[str, Any]):
    if not isinstance(dict_val, dict):
        return dict_val
    res = {}
    for k, v in dict_val.items():
        if isinstance(v, Iterable) and (("start" in v) or ("stop" in v)):
            res[k] = slice(v["start"], v["stop"], v.get("step"))
        elif isinstance(v, dict):
            res[k] = find_slice_dict_and_change_to_slice(v)
        else:
            res[k] = v

    return res


def is_time_combo(vals: dict) -> bool:
    return isinstance(vals, dict) and (
        "year" in vals or "month" in vals or "day" in vals or "hour" in vals
    )


def is_nondecreasing(values):
    return np.all(np.diff(values) >= 0)


def get_indexer_type(col):
    is_slice = False
    is_number = False
    for item in col:
        if isinstance(item, slice):
            is_slice = True
            if is_number:
                return "combo"

        if isinstance(item, Number):
            is_number = True
            if is_slice:
                return "combo"

    return "slice" if is_slice else "numeric"


def geokube_properties_to_xarray_attrs(properties, cell_methods=None):
    attrs = properties.copy()
    if "cf_name" in attrs:
        attrs["standard_name"] = attrs.pop("cf_name")
    if "description" in attrs:
        attrs["long_name"] = attrs.pop("description")
    if cell_methods:
        attrs["cell_methods"] = cell_methods.encode_for_netcdf()
    return attrs


def is_between(data, lower_bound=None, upper_bound=None):
    if lower_bound is not None:
        if upper_bound is not None:
            return (data >= lower_bound) & (data <= upper_bound)
        else:
            return data >= lower_bound
    else:
        if upper_bound is not None:
            return data <= upper_bound
        else:
            return np.full_like(data, fill_value=True, dtype=np.bool_)



import geokube.backend
from geokube import version

open_dataset = geokube.backend.open_dataset
open_datacube = geokube.backend.open_datacube

from geokube.core.coord_system import (
    AlbersEqualArea,
    GeogCS,
    Geostationary,
    LambertAzimuthalEqualArea,
    LambertConformal,
    Mercator,
    Orthographic,
    RegularLatLon,
    RotatedGeogCS,
    Stereographic,
    TransverseMercator,
    VerticalPerspective,
    CurvilinearGrid,
)

__version__ = version.__version__


import os
import timeit

import numpy as np
import pytest
from dask.delayed import Delayed

from geokube.core.field import Field
from geokube.backend import open_dataset, open_datacube

from tests import *


def test_open_datacube_and_single_time_sel():
    kube = open_datacube(
        os.path.join("tests", "resources", "rlat-rlon-tmin2m.nc")
    )
    res = kube["air_temperature"].sel(time="2007-05-02")

    assert isinstance(res, Field)
    assert np.all(
        res["time"].values.astype("datetime64[D]")
        == np.datetime64("2007-05-02")
    )


def test_open_dataset_with_load_files_on_persistance_set_to_false():
    clear_test_res()
    dset = open_dataset(
        os.path.join(
            "tests", "resources", "era5-single-levels-reanalysis_*.nc"
        ),
        os.path.join(
            "tests", "resources", "era5-single-levels-reanalysis_{var}.nc"
        ),
        load_files_on_persistance=False,
    )
    assert dset.cubes is None
    dset.persist(RES_DIR)
    assert len(os.listdir(RES_DIR)) == 2
    clear_test_res()


def test_open_geotif():
    open_datacube(
        path=os.path.join("tests", "resources", "geotiff_sample.tif")
    )



import logging
import logging.handlers
from typing import Any, Optional


class LogObject:
    def __init__(
        self,
        username: Optional[str] = "<unknown>",
        request_id: Optional[str] = "<unknown>",
    ):
        self.username = username
        self.request_id = request_id

    @classmethod
    def new(cls):
        return LogObject()

    @property
    def dict_repr(self):
        return {"user_id": self.username, "request_id": self.request_id}

    def __repr__(self):
        return str(self.dict_repr)

    def __str__(self):
        return str(self.__repr__())


class SystemLogFilter(logging.Filter):
    def filter(self, record):
        if not hasattr(record, "user_id"):
            record.user_id = "<unknown>"
        if not hasattr(record, "request_id"):
            record.request_id = "<unknown>"
        return True


def configure_logger(
    logger_name: str,
    log_path: str = "./geokube",
    debug_level: Any = logging.DEBUG,
) -> logging.Logger:
    logger = logging.getLogger(logger_name)

    # if logger.hasHandlers():
    #     return logger

    if not log_path.endswith("log"):
        log_path = log_path + ".log"
    formatter = logging.Formatter(
        "%(asctime)s:%(msecs)s %(levelname)s %(user_id)s %(request_id)s"
        " %(module)s %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
    )
    logger.setLevel(debug_level)
    fh = logging.handlers.RotatingFileHandler(log_path, "a", 30000, 30)
    fh.setLevel(debug_level)

    ch = logging.StreamHandler()
    ch.setLevel(debug_level)

    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addFilter(SystemLogFilter())
    logger.addHandler(fh)
    logger.addHandler(ch)
    return logger



from __future__ import annotations
import warnings

import os
import json
import uuid
import tempfile
import shutil
from collections.abc import Callable, Mapping, Sequence
from numbers import Number
from typing import Any, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from dask.delayed import Delayed
from zipfile import ZipFile

from ..utils.decorators import geokube_logging
from ..utils import util_methods
from ..utils.hcube_logger import HCubeLogger
from .enums import MethodType, RegridMethod
from .errs import EmptyDataError
from .axis import Axis
from .datacube import DataCube
from .domain import Domain
from .field import Field


class Dataset:
    __slots__ = (
        "__data",
        "__metadata",
        "__attrs",
        "__cube_idx",
        "__load_files_on_persistance",
    )

    _LOG = HCubeLogger(name="Dataset")

    FIELD_COL = "fields"
    DATACUBE_COL = "datacube"
    FILES_COL = "files"

    def __init__(
        self,
        hcubes: Mapping[tuple[str, ...], DataCube] | pd.DataFrame,
        attrs: Sequence[str] = None,
        metadata: Mapping[str, str] | None = None,
        load_files_on_persistance: None | bool = True,
    ) -> None:
        # NOTE: to support Dataset operations
        # for not-netcdf files
        self.__load_files_on_persistance = load_files_on_persistance
        # TODO: Make `attrs` capable of taking `np.ndarray`.
        if attrs is None:
            attrs = []
        if isinstance(hcubes, Mapping):
            self.__data = pd.Series(hcubes).reset_index()
            self.__data.columns = attrs + [self.DATACUBE_COL]
            self.__attrs = list(attrs)
        elif isinstance(hcubes, pd.DataFrame):
            self.__data = hcubes
            self.__attrs = [
                attr
                for attr in hcubes.columns
                if attr != self.DATACUBE_COL
                and attr != self.FILES_COL
                and attr != self.FIELD_COL
            ]
        else:
            raise TypeError("'hcubes' must be mapping or pandas DataFrame")

        self.__data[self.FIELD_COL] = [
            None
            if isinstance(hcube, Delayed) or hcube is None
            else list(hcube._fields.keys())
            for hcube in self.__data[self.DATACUBE_COL].to_numpy().flat
        ]
        self.__cube_idx = len(self.__attrs) + 1
        self.__metadata = dict(metadata) if metadata is not None else {}

    def __getitem__(self, key: Union[str, Tuple[str]]) -> Dataset:
        # TODO: Check if `.copy()` is necessary here.
        data = self.__data.iloc[:, : self.__cube_idx].copy()
        key = {key} if isinstance(key, str) else set(key)
        data[self.DATACUBE_COL] = [
            None
            if isinstance(hcube, Delayed)
            else Dataset._get_eligible_fields_for_datacube(hcube, key)
            for hcube in self.__data[self.DATACUBE_COL].to_numpy().flat
        ]
        dset = Dataset(
            attrs=self.__attrs, hcubes=data, metadata=self.__metadata
        )
        return dset._drop_empty()

    def __len__(self):
        return len(self.__data)

    @staticmethod
    def _get_eligible_fields_for_datacube(hcube, key: set):
        return hcube[
            (key & hcube._fields.keys()) | (key & hcube._ncvar_to_name.keys())
        ]

    def geobbox(
        self,
        north=None,
        south=None,
        west=None,
        east=None,
        top=None,
        bottom=None,
    ):
        # this returns a new Dataset where each Datacube is subsetted according to the coordinates
        _copy = self.__data.copy()
        _copy[self.DATACUBE_COL] = _copy[self.DATACUBE_COL].apply(
            lambda x: x.geobbox(
                north=north,
                south=south,
                west=west,
                east=east,
                top=top,
                bottom=bottom,
            )
        )
        return Dataset(
            attrs=self.__attrs, hcubes=_copy, metadata=self.__metadata
        )

    def locations(
        self,
        latitude,
        longitude,
        vertical: Optional[List[Number]] = None,
    ):
        # this returns a new Dataset where each Datacube is subsetted according to the coordinates
        _copy = self.__data.copy()
        _copy[self.DATACUBE_COL] = _copy[self.DATACUBE_COL].apply(
            lambda x: x.locations(
                latitude=latitude, longitude=longitude, vertical=vertical
            )
        )
        return Dataset(
            attrs=self.__attrs, hcubes=_copy, metadata=self.__metadata
        )

    def sel(
        self,
        indexers: Mapping[Union[Axis, str], Any] = None,
        method: str = None,
        tolerance: Number = None,
        drop: bool = False,
        roll_if_needed: bool = True,
        **indexers_kwargs: Any,
    ) -> Dataset:  # this can be only independent variables
        _copy = self.__data.copy()
        _copy[self.DATACUBE_COL] = _copy[self.DATACUBE_COL].apply(
            lambda x: x.sel(
                indexers=indexers,
                method=method,
                tolerance=tolerance,
                drop=drop,
                roll_if_needed=roll_if_needed,
                **indexers_kwargs,
            )
        )
        return Dataset(
            attrs=self.__attrs, hcubes=_copy, metadata=self.__metadata
        )

    def resample(
        self,
        operator: Union[Callable, MethodType, str],
        frequency: str,
        **resample_kwargs,
    ) -> Dataset:
        _copy = self.__data.copy()
        _copy[self.DATACUBE_COL] = _copy[self.DATACUBE_COL].apply(
            lambda x: x.resample(
                operator=operator, frequency=frequency, **resample_kwargs
            )
        )
        return Dataset(
            attrs=self.__attrs, hcubes=_copy, metadata=self.__metadata
        )

    def average(self, dim: str | None = None) -> Dataset:
        _copy = self.__data.copy()
        _copy[self.DATACUBE_COL] = _copy[self.DATACUBE_COL].apply(
            lambda x: x.average(dim=dim)
        )
        return Dataset(
            attrs=self.__attrs, hcubes=_copy, metadata=self.__metadata
        )

    def regrid(
        self,
        target: Union[Domain, Field],
        method: Union[str, RegridMethod] = "bilinear",
        weights_path: Optional[str] = None,
        reuse_weights: bool = True,
    ) -> Dataset:
        _copy = self.__data.copy()
        _copy[self.DATACUBE_COL] = _copy[self.DATACUBE_COL].apply(
            lambda x: x.regrid(
                target=target,
                method=method,
                weights_path=weights_path,
                reuse_weights=reuse_weights,
            )
        )
        return Dataset(
            attrs=self.__attrs, hcubes=_copy, metadata=self.__metadata
        )

    def to_regular(self) -> Dataset:
        _copy = self.__data.copy()
        _copy[self.DATACUBE_COL] = _copy[self.DATACUBE_COL].apply(
            lambda x: x.to_regular()
        )
        return Dataset(
            attrs=self.__attrs, hcubes=_copy, metadata=self.__metadata
        )

    def update_metadata(self, metadata: dict):
        self.__metadata.update(metadata)

    @property
    def data(self) -> pd.DataFrame:
        return self.__data

    @property
    def metadata(self) -> dict[str, str]:
        return self.__metadata

    @property
    def cubes(self) -> List[DataCube]:
        if self.__load_files_on_persistance:
            return self.__data[self.DATACUBE_COL].tolist()
        else:
            return None

    def to_dict(self, unique_values=False) -> dict:
        res = self.__data.drop(
            labels=Dataset.FILES_COL, inplace=False, axis=1
        ).apply(
            Dataset._row_to_dict,
            attrs=self.__attrs,
            unique_values=unique_values,
            axis=1,
        )
        return list(res)

    @staticmethod
    def _row_to_dict(row, attrs, unique_values):
        return {
            "datacube": None
            if isinstance(row[Dataset.DATACUBE_COL], Delayed)
            else row[Dataset.DATACUBE_COL].to_dict(unique_values),
            "attributes": {attr_name: row[attr_name] for attr_name in attrs},
        }

    def _drop_empty(self) -> Dataset:
        mask = self.__data[self.FIELD_COL].astype(dtype=np.bool_)
        data = self.__data.loc[mask, : self.DATACUBE_COL]
        data.index = np.arange(len(data))
        return Dataset(
            attrs=self.__attrs, hcubes=data, metadata=self.__metadata
        )

    @geokube_logging
    def filter(
        self, indexers: Optional[Mapping[str, str]] = None, **indexers_kwargs
    ) -> Dataset:
        if indexers is None:
            params = indexers_kwargs
        else:
            if intersect := sorted(indexers.keys() & indexers_kwargs.keys()):
                raise ValueError(
                    "'indexers' and 'indexers_kwargs' have common parameters:"
                    " {intersect}"
                )
            params = {**indexers, **indexers_kwargs}

        if not (idx := params.keys()) <= (attrs := set(self.__attrs)):
            # TODO: Make better message.
            raise ValueError(
                f"'filter' cannot use the argument(s): {sorted(idx - attrs)}"
            )

        mask = np.full(shape=len(self.__data), fill_value=True, dtype=np.bool_)
        for param_name, param_value in params.items():
            mask &= np.in1d(self.__data[param_name], param_value)
        data = self.__data.loc[mask, : self.DATACUBE_COL]
        data.index = np.arange(len(data))

        return Dataset(
            attrs=self.__attrs,
            hcubes=data,
            metadata=self.__metadata,
            load_files_on_persistance=self.__load_files_on_persistance,
        )

    def apply(
        self,
        func: Union[Callable[[DataCube], DataCube], property, str],
        drop_empty: Optional[bool] = False,
        **kwargs,
    ) -> Dataset:
        data = self.__data.iloc[:, : self.__cube_idx].copy()
        data[self.DATACUBE_COL] = [
            _apply(hcube, func, **kwargs)
            for hcube in self.__data[self.DATACUBE_COL].to_numpy().flat
        ]
        data.index = np.arange(len(data))
        dset = Dataset(
            attrs=self.__attrs, hcubes=data, metadata=self.__metadata
        )
        return dset._drop_empty() if drop_empty else dset

    def persist(self, path=None, zip_if_many=False) -> str:
        if path is None:
            path = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        if not os.path.exists(path):
            os.makedirs(path)

        paths = self.data.apply(self._persist_datacube, path=path, axis=1)
        # NOTE: omit None paths -- results of empty DataCube persistance
        paths = paths[~paths.isna()]
        if len(paths) == 0:
            warnings.warn(
                "No files were created while geokube.Dataset persisting!"
            )
            return None
        elif len(paths) == 1:
            return paths.iloc[0]
        if zip_if_many:
            path = os.path.join(path, f"{str(uuid.uuid4())}.zip")
            with ZipFile(path, "w") as archive:
                for file in paths:
                    archive.write(file, arcname=os.path.basename(file))
            for file in paths:
                os.remove(file)
        return path

    @property
    def nbytes(self) -> int:
        if any(isinstance(cube, Delayed) for cube in self.cubes):
            return sum(
                sum(os.path.getsize(f) for f in files)
                for files in self.__data[Dataset.FILES_COL]
            )
        return sum(cube.nbytes for cube in self.cubes)

    def _persist_datacube(self, dataframe_item, path):
        if self.__load_files_on_persistance:
            dcube = dataframe_item[self.DATACUBE_COL]
            attr_str = self._form_attr_str(dataframe_item)
            if isinstance(dcube, Delayed):
                dcube = dcube.compute()
            try:
                return dcube.persist(os.path.join(path, f"{attr_str}.nc"))
            except EmptyDataError:
                self._LOG.warn(f"Skipping empty Dataset item!")
                return None
        else:
            attr_str = self._form_attr_str(dataframe_item)
            if len(dataframe_item[Dataset.FILES_COL]) > 1:
                raise ValueError(
                    "Too many files! Copying source files is supported for"
                    " `1` but provided"
                    f" `{len(dataframe_item[Dataset.FILES_COL])}`"
                )
            for file in dataframe_item[Dataset.FILES_COL]:
                dst_path = os.path.join(
                    path,
                    self._convert_attributes_to_file_name(attr_str, file),
                )
                shutil.copyfile(file, dst_path)
                return dst_path

    def _form_attr_str(self, dataframe_item):
        return "-".join(
            [
                f"{attr_name}={dataframe_item[attr_name]}"
                for attr_name in self.__attrs
            ]
        )

    def _convert_attributes_to_file_name(self, attr_str, file):
        return f"{attr_str}-{os.path.basename(file)}"


def _apply(
    hcube: DataCube,
    func: Union[Callable[[DataCube], DataCube], property, str],
    **kwargs,
) -> DataCube:
    if isinstance(func, str):
        func = getattr(DataCube, func)
    if callable(func):
        return func(hcube, **kwargs)
    if isinstance(func, property):
        if kwargs:
            raise ValueError(
                "'func' is a property and cannot accept additional arguments"
            )
        return func.fget(hcube)
    raise TypeError(
        "'func' must be callable or str which represents the name of a "
        "callable member of DataCube class"
    )



import uuid
from collections import OrderedDict
from functools import lru_cache, partial
from html import escape

import pkg_resources
from xarray.core.formatting import inline_variable_array_repr, short_data_repr
from xarray.core.options import _get_boolean_with_default

from ..core.axis import Axis

STATIC_FILES = ("static/html/icons-svg-inline.html", "static/css/style.css")


@lru_cache(None)
def _load_static_files():
    """Lazily load the resource files into memory the first time they are needed"""
    return [
        pkg_resources.resource_string("geokube", fname).decode("utf8")
        for fname in STATIC_FILES
    ]


def short_data_repr_html(array):
    """Format "data" for DataArray and Variable."""
    internal_data = getattr(array, "variable", array)._data
    if hasattr(internal_data, "_repr_html_"):
        return internal_data._repr_html_()
    text = escape(short_data_repr(array))
    return f"<pre>{text}</pre>"


def format_dims(dims, coord_names):
    if not dims:
        return ""

    dim_css_map = {
        k: " class='xr-has-index'" if k in coord_names else ""
        for k, v in dims.items()
    }

    dims_li = "".join(
        f"<li><span{dim_css_map[dim]}>{escape(str(dim))}</span>: {size}</li>"
        for dim, size in dims.items()
    )

    return f"<ul class='xr-dim-list'>{dims_li}</ul>"


def summarize_attrs(attrs):
    attrs_dl = "".join(
        f"<dt><span>{escape(str(k))} :</span></dt><dd>{escape(str(v))}</dd>"
        for k, v in attrs.items()
    )

    return f"<dl class='xr-attrs'>{attrs_dl}</dl>"


def _icon(icon_name):
    # icon_name should be defined in xarray/static/html/icon-svg-inline.html
    return (
        "<svg class='icon xr-{0}'><use xlink:href='#{0}'></use></svg>".format(
            icon_name
        )
    )


def _summarize_coord_multiindex(name, coord):
    preview = f"({', '.join(escape(l) for l in coord.level_names)})"
    return summarize_variable(
        name, coord, is_index=True, dtype="MultiIndex", preview=preview
    )


def summarize_coord(name, var):
    is_index = name in var.dims
    if is_index:
        coord = var.variable.to_index_variable()
        if coord.level_names is not None:
            coords = {name: _summarize_coord_multiindex(name, coord)}
            for lname in coord.level_names:
                var = coord.get_level_variable(lname)
                coords[lname] = summarize_variable(lname, var)
            return coords

    return {name: summarize_variable(name, var, is_index)}


def summarize_coords(variables):
    coords = {}
    for k, v in variables.items():
        coords.update(**summarize_coord(k, v))

    vars_li = "".join(
        f"<li class='xr-var-item'>{v}</li>" for v in coords.values()
    )

    return f"<ul class='xr-var-list'>{vars_li}</ul>"


def summarize_variable(name, var, is_index=False, dtype=None, preview=None):
    variable = var.variable if hasattr(var, "variable") else var

    cssclass_idx = " class='xr-has-index'" if is_index else ""
    dims_str = f"({', '.join(escape(dim) for dim in var.dims)})"
    name = escape(str(name))
    dtype = dtype or escape(str(var.dtype))

    # "unique" ids required to expand/collapse subsections
    attrs_id = "attrs-" + str(uuid.uuid4())
    data_id = "data-" + str(uuid.uuid4())
    disabled = "" if len(var.attrs) else "disabled"

    preview = preview or escape(inline_variable_array_repr(variable, 35))
    attrs_ul = summarize_attrs(var.attrs)
    data_repr = short_data_repr_html(variable)

    attrs_icon = _icon("icon-file-text2")
    data_icon = _icon("icon-database")

    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
        f"<div class='xr-var-dims'>{dims_str}</div>"
        f"<div class='xr-var-dtype'>{dtype}</div>"
        f"<div class='xr-var-preview xr-preview'>{preview}</div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox' {disabled}>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<input id='{data_id}' class='xr-var-data-in' type='checkbox'>"
        f"<label for='{data_id}' title='Show/Hide data repr'>"
        f"{data_icon}</label>"
        f"<div class='xr-var-attrs'>{attrs_ul}</div>"
        f"<div class='xr-var-data'>{data_repr}</div>"
    )


def summarize_vars(variables):
    vars_li = "".join(
        f"<li class='xr-var-item'>{summarize_variable(k, v)}</li>"
        for k, v in variables.items()
    )

    return f"<ul class='xr-var-list'>{vars_li}</ul>"


def collapsible_section(
    name,
    inline_details="",
    details="",
    n_items=None,
    enabled=True,
    collapsed=False,
):
    # "unique" id to expand/collapse the section
    data_id = "section-" + str(uuid.uuid4())

    has_items = n_items is not None and n_items
    n_items_span = "" if n_items is None else f" <span>({n_items})</span>"
    enabled = "" if enabled and has_items else "disabled"
    collapsed = "" if collapsed or not has_items else "checked"
    tip = " title='Expand/collapse section'" if enabled else ""

    return (
        f"<input id='{data_id}' class='xr-section-summary-in' "
        f"type='checkbox' {enabled} {collapsed}>"
        f"<label for='{data_id}' class='xr-section-summary' {tip}>"
        f"{name}:{n_items_span}</label>"
        f"<div class='xr-section-inline-details'>{inline_details}</div>"
        f"<div class='xr-section-details'>{details}</div>"
    )


def _mapping_section(
    mapping,
    name,
    details_func,
    max_items_collapse,
    expand_option_name,
    enabled=True,
):
    n_items = len(mapping)
    expanded = _get_boolean_with_default(
        expand_option_name, n_items < max_items_collapse
    )
    collapsed = not expanded

    return collapsible_section(
        name,
        details=details_func(mapping),
        n_items=n_items,
        enabled=enabled,
        collapsed=collapsed,
    )


def dim_section(obj):
    dim_list = format_dims(obj.dims, list(obj.coords))

    return collapsible_section(
        "Dimensions", inline_details=dim_list, enabled=False, collapsed=True
    )


def array_section(obj):
    # "unique" id to expand/collapse the section
    data_id = "section-" + str(uuid.uuid4())
    collapsed = (
        "checked"
        if _get_boolean_with_default("display_expand_data", default=True)
        else ""
    )
    variable = getattr(obj, "variable", obj)
    preview = escape(inline_variable_array_repr(variable, max_width=70))
    data_repr = short_data_repr_html(obj)
    data_icon = _icon("icon-database")

    return (
        f"<div class='xr-array-wrap'><input id='{data_id}' class='xr-array-in'"
        f" type='checkbox' {collapsed}><label for='{data_id}' title='Show/hide"
        f" data repr'>{data_icon}</label><div class='xr-array-preview"
        f" xr-preview'><span>{preview}</span></div><div"
        f" class='xr-array-data'>{data_repr}</div></div>"
    )


coord_section = partial(
    _mapping_section,
    name="Coordinates",
    details_func=summarize_coords,
    max_items_collapse=25,
    expand_option_name="display_expand_coords",
)


datavar_section = partial(
    _mapping_section,
    name="Data variables",
    details_func=summarize_vars,
    max_items_collapse=15,
    expand_option_name="display_expand_data_vars",
)


attr_section = partial(
    _mapping_section,
    name="Attributes",
    details_func=summarize_attrs,
    max_items_collapse=10,
    expand_option_name="display_expand_attrs",
)


def _obj_repr(obj, header_components, sections):
    """Return HTML repr of an xarray object.
    If CSS is not injected (untrusted notebook), fallback to the plain text repr.
    """
    header = (
        f"<div class='xr-header'>{''.join(h for h in header_components)}</div>"
    )
    sections = "".join(
        f"<li class='xr-section-item'>{s}</li>" for s in sections
    )

    icons_svg, css_style = _load_static_files()
    return (
        "<div>"
        f"{icons_svg}<style>{css_style}</style>"
        f"<pre class='xr-text-repr-fallback'>{escape(repr(obj))}</pre>"
        "<div class='xr-wrap' style='display:none'>"
        f"{header}"
        f"<ul class='xr-sections'>{sections}</ul>"
        "</div>"
        "</div>"
    )


def array_repr(arr):
    dims = OrderedDict((k, v) for k, v in zip(arr.dims, arr.shape))

    obj_type = "geokube.{}".format(type(arr).__name__)
    arr_name = f"'{arr.name}'" if getattr(arr, "name", None) else ""
    coord_names = list(arr.coords) if hasattr(arr, "coords") else []

    header_components = [
        f"<div class='xr-obj-type'>{obj_type}</div>",
        f"<div class='xr-array-name'>{arr_name}</div>",
        format_dims(dims, coord_names),
    ]

    sections = [array_section(arr)]

    if hasattr(arr, "coords"):
        sections.append(coord_section(arr.coords))

    sections.append(attr_section(arr.attrs))

    return _obj_repr(arr, header_components, sections)


def dataset_repr(ds):
    obj_type = "geokube.{}".format(type(ds).__name__)

    header_components = [f"<div class='xr-obj-type'>{escape(obj_type)}</div>"]

    sections = [
        dim_section(ds),
        coord_section(ds.coords),
        datavar_section(ds.data_vars),
        attr_section(ds.attrs),
    ]

    return _obj_repr(ds, header_components, sections)



import cf_units as cf
import pytest

from geokube.core.unit import Unit


def test_unit_proper_attr_setting():
    u = Unit("not_existing")
    assert u.is_unknown
    assert str(u) == "not_existing"
    assert u._unit == cf.Unit(None)

    u = Unit("m")
    assert str(u) == "m"
    assert not u.is_time_reference()

    u = Unit("hours since 1970-01-01 00:00:00", "gregorian")
    assert u.is_time_reference()
